from pathlib import Path
from django.core.management.utils import get_random_secret_key
import os
import sys
from configparser import ConfigParser
import importlib.metadata
import bomiot
from bomiot.server.server.pkgcheck import pkg_check, cwd_check, ignore_pkg, ignore_cwd
import importlib.util
from os import listdir
from os.path import join, isdir, exists, isfile


BASE_DIR = join(Path(bomiot.__file__).resolve().parent, 'server')

WORKING_SPACE_CONFIG = ConfigParser()
WORKING_SPACE_CONFIG.read(join(BASE_DIR, 'workspace.ini'), encoding='utf-8')
WORKING_SPACE = WORKING_SPACE_CONFIG.get('space', 'name', fallback='Create your working space first')
sys.path.insert(0, WORKING_SPACE)

CONFIG = ConfigParser()
setup_ini_path = join(WORKING_SPACE, 'setup.ini')
CONFIG.read(setup_ini_path, encoding='utf-8')
PROJECT_NAME = CONFIG.get('project', 'name', fallback='bomiot')

SECRET_KEY = get_random_secret_key()

DEBUG = True

ALLOWED_HOSTS = ['*', '127.0.0.1', 'localhost']
AUTH_USER_MODEL = "core.User"

INSTALLED_APPS = [
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',
    'corsheaders',
    'django_filters',
    'rest_framework',
    'django_apscheduler',
    'bomiot.server.core',
]

all_packages = [dist.metadata['Name'] for dist in importlib.metadata.distributions()]
res_pkg_list = list(set([name.lower() for name in all_packages]).difference(set(ignore_pkg())))
pkg_squared = list(map(lambda data: pkg_check(data), res_pkg_list))
filtered_pkg_squared = list(filter(lambda x: x is not None, pkg_squared))

current_path = list(set([p for p in listdir(WORKING_SPACE) if isdir(p)]).difference(set(ignore_cwd())))
cur_squared = list(map(lambda data: cwd_check(data), current_path))
filtered_current_path = list(filter(lambda y: y is not None, cur_squared))

def load_dynamic_apps() -> None:
    global INSTALLED_APPS
    load_apps_from_project()
    load_apps_from_working_space()
    load_apps_from_packages()
    

def load_apps_from_packages() -> None:
    global INSTALLED_APPS
    all_packages = [dist.metadata['Name'] for dist in importlib.metadata.distributions()]
    res_pkg_list = list(set([name.lower() for name in all_packages]).difference(set(ignore_pkg())))
    for module in res_pkg_list:
        spec = importlib.util.find_spec(module)
        if not spec:
            continue
        module_path = spec.origin
        if not module_path:
            continue
        module_dir = Path(module_path).resolve().parent
        config_path = join(module_dir, 'bomiotconf.ini')
        pkg_config = ConfigParser()
        if pkg_config.read(config_path, encoding='utf-8'):
            app_mode = pkg_config.get('mode', 'name', fallback='plugins')
            if app_mode == 'plugins':
                apps_py_path = join(module_dir, 'apps.py')
                if isfile(apps_py_path):
                    app_name = module
                    if app_name not in INSTALLED_APPS:
                        INSTALLED_APPS.append(app_name)
                    else:
                        continue
                else:
                    continue
            else:
                continue
        else:
            continue

def load_apps_from_working_space() -> None:
    global INSTALLED_APPS
    current_path = [
        p for p in listdir(WORKING_SPACE) 
        if isdir(join(WORKING_SPACE, p)) and p not in ignore_cwd()
    ]
    for module_name in current_path:
        module_dir = join(WORKING_SPACE, module_name)
        config_path = join(module_dir, 'bomiotconf.ini')
        app_mode_config = ConfigParser()
        if not app_mode_config.read(config_path, encoding='utf-8'):
            continue
        app_mode = app_mode_config.get('mode', 'name')
        if app_mode == 'plugins':
            apps_py_path = join(module_dir, 'apps.py')
            if isfile(apps_py_path):
                if module_name not in INSTALLED_APPS:
                    INSTALLED_APPS.append(module_name)
                else:
                    continue
            else:
                continue
        else:
            continue

def load_apps_from_project() -> None:
    global INSTALLED_APPS, PROJECT_NAME
    if not PROJECT_NAME or PROJECT_NAME not in [p for p in listdir(WORKING_SPACE) if isdir(p)]:
        return
    project_path = join(WORKING_SPACE, PROJECT_NAME)
    app_dirs = [
        app for app in listdir(project_path) 
        if isdir(join(project_path, app))
    ]
    for app in app_dirs:
        apps_py_path = join(project_path, app, 'apps.py')
        if isfile(apps_py_path):
            app_name = f'{PROJECT_NAME}.{app}'
            if app_name not in INSTALLED_APPS:
                INSTALLED_APPS.append(app_name)
            else:
                continue
        else:
            continue

load_dynamic_apps()

MIDDLEWARE = [
    'django.middleware.gzip.GZipMiddleware',
    'django.middleware.security.SecurityMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.locale.LocaleMiddleware',
    'corsheaders.middleware.CorsMiddleware',
    'django.middleware.common.CommonMiddleware',
    # 'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    # 'django.middleware.clickjacking.XFrameOptionsMiddleware',
]

ROOT_URLCONF = 'bomiot.server.server.urls'

TEMPLATES_PATH = join(WORKING_SPACE, PROJECT_NAME)

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [TEMPLATES_PATH],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
                'django.template.context_processors.i18n',
            ],
        },
    },
]

WSGI_APPLICATION = 'bomiot.server.server.wsgi.application'


# Database
# https://docs.djangoproject.com/en/4.2/ref/settings/#databases
BASE_DB_TABLE = 'bomiot'

DATABASE_MAP = {
    'sqlite': 'django.db.backends.sqlite3',
    'mysql': 'django.db.backends.mysql',
    'postgresql': 'django.db.backends.postgresql_psycopg2',
    'oracle': 'django.db.backends.oracle',
}

db_engine = CONFIG.get('database', 'engine', fallback='sqlite')
if db_engine == 'sqlite':
    DB_DIR = join(WORKING_SPACE, 'dbs')
    exists(DB_DIR) or os.makedirs(DB_DIR)
    DB_PATH = join(DB_DIR, 'db.sqlite3')
    DATABASES = {
        'default': {
            'ENGINE': 'django.db.backends.sqlite3',
            'NAME': DB_PATH,
            'CONN_MAX_AGE': 60,
            'OPTIONS': {
                'timeout': 60,
            }
        }
    }
elif db_engine == 'mysql':
    DATABASES = {
        'default': {
            'ENGINE': DATABASE_MAP[CONFIG['database']['engine']],
            'NAME': CONFIG['database']['name'],
            'USER': CONFIG['database']['user'],
            'PASSWORD': CONFIG['database']['password'],
            'HOST': CONFIG['database']['host'],
            'PORT': CONFIG['database']['port'],
            'OPTIONS': {
                'charset': 'utf8mb4',
            },
        }
    }
else:
    DATABASES = {
        'default': {
            'ENGINE': DATABASE_MAP[CONFIG['database']['engine']],
            'NAME': CONFIG['database']['name'],
            'USER': CONFIG['database']['user'],
            'PASSWORD': CONFIG['database']['password'],
            'HOST': CONFIG['database']['host'],
            'PORT': CONFIG['database']['port'],
        }
    }



# Password validation
# https://docs.djangoproject.com/en/4.2/ref/settings/#auth-password-validators

AUTH_PASSWORD_VALIDATORS = [
    {
        'NAME': 'django.contrib.auth.password_validation.UserAttributeSimilarityValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.MinimumLengthValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.CommonPasswordValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.NumericPasswordValidator',
    },
]


# Internationalization
# https://docs.djangoproject.com/en/4.2/topics/i18n/
if PROJECT_NAME == 'bomiot':
    LANGUAGE_DIR = join(BASE_DIR, 'language').replace('\\', '/')
else:
    LANGUAGE_DIR = join(join(WORKING_SPACE, PROJECT_NAME), 'language').replace('\\', '/')
if exists(join(WORKING_SPACE, PROJECT_NAME)):
    exists(LANGUAGE_DIR) or os.makedirs(LANGUAGE_DIR)

TIME_ZONE = CONFIG.getint('local', 'time_zone', fallback='UTC')

USE_I18N = True

USE_L10N = True  # (localization)

USE_TZ = False


# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/4.2/howto/static-files/

STATIC_URL = 'static/'
if PROJECT_NAME == 'bomiot':
    STATIC_ROOT = join(BASE_DIR, 'bomiot_static').replace('\\', '/')
    STATICFILES_DIRS = [
        join(BASE_DIR, "static").replace('\\', '/'),
    ]
else:
    STATIC_ROOT = join(join(WORKING_SPACE, PROJECT_NAME), 'bomiot_static').replace('\\', '/')
    STATICFILES_DIRS = [
        join(join(WORKING_SPACE, PROJECT_NAME), 'static').replace('\\', '/')
    ]

MEDIA_URL = 'media/'
if PROJECT_NAME == 'bomiot':
    MEDIA_ROOT = join(BASE_DIR, 'media').replace('\\', '/')
else:
    MEDIA_ROOT = join(join(WORKING_SPACE, PROJECT_NAME), 'media').replace('\\', '/')
if exists(join(WORKING_SPACE, PROJECT_NAME)):
    exists(MEDIA_ROOT) or os.makedirs(MEDIA_ROOT)


# Default primary key field type
# https://docs.djangoproject.com/en/4.2/ref/settings/#default-auto-field

DEFAULT_AUTO_FIELD = 'django.db.models.BigAutoField'

CORS_ALLOW_CREDENTIALS = True
CORS_ORIGIN_ALLOW_ALL = True
CORS_ORIGIN_WHITELIST = ()

CORS_ALLOW_METHODS = (
    'DELETE',
    'GET',
    'OPTIONS',
    'PATCH',
    'POST',
    'PUT',
    'VIEW',
)

CORS_ALLOW_HEADERS = (
    'accept',
    'accept-encoding',
    'authorization',
    'content-type',
    'dnt',
    'origin',
    'user-agent',
    'x-csrftoken',
    'x-requested-with',
    'token',
    'language',
    'operator',
    'device',
    'app-id',
    'event-sign'
)

CORS_EXPOSE_HEADERS = [
    "expire"
]

X_FRAME_OPTIONS = 'SAMEORIGIN'


LOG_PATH = join(WORKING_SPACE, 'logs')
exists(LOG_PATH) or os.makedirs(LOG_PATH)
SERVER_LOGS_FILE = join(LOG_PATH, 'server.log')
ERROR_LOGS_FILE = join(LOG_PATH, 'error.log')
STANDARD_LOG_FORMAT = (
    "[%(asctime)s][%(name)s.%(funcName)s():%(lineno)d] [%(levelname)s] %(message)s"
)
CONSOLE_LOG_FORMAT = (
    "[%(asctime)s][%(name)s.%(funcName)s():%(lineno)d] [%(levelname)s] %(message)s"
)
LOGGING = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {
        "standard": {"format": STANDARD_LOG_FORMAT},
        "console": {
            "format": CONSOLE_LOG_FORMAT,
            "datefmt": "%Y-%m-%d %H:%M:%S",
        },
        "file": {
            "format": CONSOLE_LOG_FORMAT,
            "datefmt": "%Y-%m-%d %H:%M:%S",
        },
    },
    "handlers": {
        "file": {
            "level": "INFO",
            "class": "logging.handlers.RotatingFileHandler",
            "filename": SERVER_LOGS_FILE,
            "maxBytes": 1024 * 1024 * 100,  # 100 MB
            "backupCount": 5,
            "formatter": "standard",
            "encoding": "utf-8",
        },
        "error": {
            "level": "ERROR",
            "class": "logging.handlers.RotatingFileHandler",
            "filename": ERROR_LOGS_FILE,
            "maxBytes": 1024 * 1024 * 100,  # 100 MB
            "backupCount": 3,
            "formatter": "standard",
            "encoding": "utf-8",
        },
        "console": {
            "level": "INFO",
            "class": "logging.StreamHandler",
            "formatter": "console",
        },

    },
    "loggers": {
        "": {
            "handlers": ["console", "error", "file"],
            "level": "INFO",
        },
        'apscheduler': {
            'handlers': ['console'],
            'level': 'WARNING',
            'propagate': False,
        },
        "django": {
            "handlers": ["console", "error", "file"],
            "level": "INFO",
            "propagate": False,
        },
        'django.db.backends': {
            'handlers': ["console", "error", "file"],
            'propagate': False,
            'level': "INFO"
        }
    },
}

REST_FRAMEWORK = {
    # AttributeError: 'AutoSchema' object has no attribute 'get_link'
    # DEFAULT SET:
    'DEFAULT_SCHEMA_CLASS': 'rest_framework.schemas.openapi.AutoSchema',
    # EXCEPTION:
    'EXCEPTION_HANDLER': 'bomiot.server.core.my_exceptions.custom_exception_handler',
    # Base API policies:
    'DEFAULT_RENDERER_CLASSES': [
        'rest_framework.renderers.JSONRenderer',
        'rest_framework_csv.renderers.CSVRenderer',
        #'rest_framework.renderers.BrowsableAPIRenderer',
    ],
    'DEFAULT_PARSER_CLASSES': [
        'rest_framework.parsers.JSONParser',
        'rest_framework.parsers.FormParser',
        'rest_framework.parsers.MultiPartParser'
    ],
    'DEFAULT_AUTHENTICATION_CLASSES': ['bomiot.server.core.auth.CoreAuthentication', ],
    'DEFAULT_PERMISSION_CLASSES': ["bomiot.server.core.permission.CorePermission", ],
    'DEFAULT_THROTTLE_CLASSES': ['bomiot.server.core.throttle.CoreThrottle', ],
    # 'DEFAULT_THROTTLE_RATES': ['utils.throttle.VisitThrottle', ],
    'DEFAULT_CONTENT_NEGOTIATION_CLASS': 'rest_framework.negotiation.DefaultContentNegotiation',
    'DEFAULT_METADATA_CLASS': 'rest_framework.metadata.SimpleMetadata',
    'DEFAULT_VERSIONING_CLASS': None,
    'DEFAULT_PAGINATION_CLASS': 'bomiot.server.core.page.CorePageNumberPagination',
    # 'PAGE_SIZE': 1,  # 默认 None
    'DEFAULT_FILTER_BACKENDS': [
        'django_filters.rest_framework.DjangoFilterBackend',
        # 'django_filters.rest_framework.backends.DjangoFilterBackend',
    ],
    'SEARCH_PARAM': 'search',
    'ORDERING_PARAM': 'ordering',
    'NUM_PROXIES': None,
    # Versioning:
    'DEFAULT_VERSION': None,
    'ALLOWED_VERSIONS': None,
    'VERSION_PARAM': 'version',
    # Authentication:
    'UNAUTHENTICATED_USER': 'django.contrib.auth.models.AnonymousUser',
    'UNAUTHENTICATED_TOKEN': None,
    # View configuration:
    'VIEW_NAME_FUNCTION': 'rest_framework.views.get_view_name',
    'VIEW_DESCRIPTION_FUNCTION': 'rest_framework.views.get_view_description',
    'NON_FIELD_ERRORS_KEY': 'non_field_errors',
    # Testing
    'TEST_REQUEST_RENDERER_CLASSES': [
        'rest_framework.renderers.MultiPartRenderer',
        'rest_framework.renderers.JSONRenderer'
    ],
    'TEST_REQUEST_DEFAULT_FORMAT': 'multipart',
    # Hyperlink settings
    'URL_FORMAT_OVERRIDE': 'format',
    'FORMAT_SUFFIX_KWARG': 'format',
    'URL_FIELD_NAME': 'url',
    # Encoding
    'UNICODE_JSON': True,
    'COMPACT_JSON': True,
    'STRICT_JSON': True,
    'COERCE_DECIMAL_TO_STRING': True,
    'UPLOADED_FILES_USE_URL': True,
    # Browseable API
    'HTML_SELECT_CUTOFF': 1000,
    'HTML_SELECT_CUTOFF_TEXT': "More than {count} items...",
    # Schemas
    'SCHEMA_COERCE_PATH_PK': True,
    'SCHEMA_COERCE_METHOD_NAMES': {
        'retrieve': 'read',
        'destroy': 'delete'
    },
}

CACHES = {
    "default": {
        "BACKEND": "django.core.cache.backends.locmem.LocMemCache",
        "LOCATION": "unique-snowflake",
    }
}

INTERNAL_IPS = [
    '127.0.0.1',
    'localhost'
]

USER_JWT_TIME = CONFIG.getint('jwt', 'user_jwt_time', fallback=1000000)
JWT_SALT = 'ds()udsjo@jlsdosjf)wjd_#(#)$'

KEY_PATH = join(WORKING_SPACE, 'auth_key.py')
KEY = ''
if exists(KEY_PATH):
    try:
        with open(KEY_PATH, 'r', encoding='utf-8') as f:
            content = f.read()
        lines = content.split('\n')
        for line in lines:
            line = line.strip()
            if line.startswith('KEY') and '=' in line:
                key_value = line.split('=', 1)[1].strip()
                if key_value.startswith("'") and key_value.endswith("'"):
                    key_value = key_value[1:-1]
                elif key_value.startswith('"') and key_value.endswith('"'):
                    key_value = key_value[1:-1]
                KEY = key_value
                break
    except Exception as e:
        print(f"Error reading auth_key.py: {e}")
        KEY = ''

ALLOCATION_SECONDS = CONFIG.getint('throttle', 'allocation_seconds', fallback=1)
THROTTLE_SECONDS = CONFIG.getint('throttle', 'throttle_seconds', fallback=10)

REQUEST_LIMIT = CONFIG.getint('request', 'limit', fallback=2)

DATA_UPLOAD_MAX_MEMORY_SIZE = None

FILE_SIZE = CONFIG.getint('file', 'file_size', fallback=102400000)
FILE_EXTENSION = CONFIG.get('file', 'file_extension', fallback='py,png,jpg,jpeg,gif,bmp,webp,txt,md,html,htm,js,css,json,xml,csv,xlsx,xls,ppt,pptx,doc,docx,pdf').replace(" ", "").split(',')

EMAIL_BACKEND = 'django.core.mail.backends.smtp.EmailBackend'
EMAIL_HOST = CONFIG.get('mail', 'email_host', fallback='')
EMAIL_PORT = CONFIG.getint('mail', 'email_port', fallback=465)
EMAIL_HOST_USER = CONFIG.get('mail', 'email_host_user', fallback='')
EMAIL_HOST_PASSWORD = CONFIG.get('mail', 'email_host_password', fallback='')
DEFAULT_FROM_EMAIL = CONFIG.get('mail', 'default_from_email', fallback='')
EMAIL_FROM = CONFIG.get('mail', 'email_from', fallback='')
EMAIL_USE_SSL = CONFIG.getboolean('mail', 'email_use_ssl', fallback=True)


import time
import psutil
import threading
import pypistats
import orjson
from datetime import datetime, timedelta
from bomiot.server.core.models import Pids, PyPi, CPU, Memory, Disk, Network
from bomiot.server.core.utils import readable_file_size
from bomiot.server.core.signal import bomiot_signals
from time import sleep

pypi_stats_list = []

class ServerManager:
    def __init__(self):
        """init server manager"""
        pass
    
    def get_cpu_info(self):
        """CPU"""
        cpu_percent = psutil.cpu_percent(interval=1)
        cpu_count = psutil.cpu_count(logical=False)
        cpu_count_logical = psutil.cpu_count(logical=True)
        cpu_freq = psutil.cpu_freq()
        cpu_info_check = CPU.objects.filter()
        if cpu_info_check.count() >= 10080:
            cpu_info_check.order_by('id').first().delete()
        cpu_data = CPU.objects.create(
                   cpu_usage=float(f"{cpu_percent:.2f}"),
                   physical_cores=int(cpu_count),
                   logical_cores=int(cpu_count_logical),
                   cpu_frequency=f"{cpu_freq.current:.2f} MHz",
                   min_cpu_frequency=f"{cpu_freq.min:.2f} MHz",
                   max_cpu_frequency=f"{cpu_freq.max:.2f} MHz"
                   )
        bomiot_signals.send(sender=CPU, msg={
            'models': 'CPU',
            'type': 'created',
            'data': {
                'id': cpu_data.id,
                'cpu_usage': float(f"{cpu_percent:.2f}"),
                'physical_cores': int(cpu_count),
                'logical_cores': int(cpu_count_logical),
                'cpu_frequency': f"{cpu_freq.current:.2f} MHz",
                'min_cpu_frequency': f"{cpu_freq.min:.2f} MHz",
                'max_cpu_frequency': f"{cpu_freq.max:.2f} MHz"
            }
        })

    def get_memory_info(self):
        """Memory"""
        memory = psutil.virtual_memory()
        swap = psutil.swap_memory()
        memory_info_check = Memory.objects.filter()
        if memory_info_check.count() >= 10080:
            memory_info_check.order_by('id').first().delete()
        memory_data = Memory.objects.create(
                      total=int(memory.total),
                      used=int(memory.used),
                      free=int(memory.free),
                      percent=float(f'{memory.percent:.2f}'),
                      swap_total=int(swap.total),
                      swap_used=int(swap.used),
                      swap_free=int(swap.free),
                      swap_percent=float(f'{swap.percent:.2f}'),
                      )
        bomiot_signals.send(sender=Memory, msg={
            'models': 'Memory',
            'type': 'created',
            'data': {
                'id': memory_data.id,
                'total': int(memory.total),
                'used': int(memory.used),
                'free': int(memory.free),
                'percent': float(f'{memory.percent:.2f}'),
                'swap_total': int(swap.total),
                'swap_used': int(swap.used),
                'swap_free': int(swap.free),
                'swap_percent': float(f'{swap.percent:.2f}')
            }
        })

    def get_disk_info(self):
        """Disk"""
        Disk.objects.all().delete()  # Clear all previous records
        partitions = psutil.disk_partitions()
        disk_list = []
        for i in range(len(partitions)):
            try:
                disk_usage = psutil.disk_usage(partitions[i].mountpoint)
                disk_detail = Disk(
                    device=partitions[i].device,
                    mountpoint=partitions[i].mountpoint,
                    total=int(disk_usage.total),
                    used=int(disk_usage.used),
                    free=int(disk_usage.free),
                    percent=float(f'{disk_usage.percent:.2f}')
                )
                bomiot_signals.send(sender=Disk, msg={
                    'models': 'Disk',
                    'type': 'created',
                    'data': {
                        'device': partitions[i].device,
                        'mountpoint': partitions[i].mountpoint,
                        'total': int(disk_usage.total),
                        'used': int(disk_usage.used),
                        'free': int(disk_usage.free),
                        'percent': float(f'{disk_usage.percent:.2f}')
                    }
                })
                disk_list.append(disk_detail)
            except PermissionError:
                print(f"{partitions[i].mountpoint}")
                continue
        if len(disk_list) > 0:
            Disk.objects.bulk_create(disk_list, batch_size=100)

    def get_network_info(self):
        """Network"""
        net_io_counters = psutil.net_io_counters()
        bytes_sent = net_io_counters.bytes_sent
        bytes_recv = net_io_counters.bytes_recv
        newtork_info_check = Network.objects.filter()
        if newtork_info_check.count() >= 10080:
            newtork_info_check.order_by('id').first().delete()
        instance = Network.objects.create(
                   bytes_sent=int(bytes_sent),
                   bytes_recv=int(bytes_recv)
                   )
        bomiot_signals.send(sender=Network, msg={
            'models': 'Network',
            'type': 'created',
            'data': {
                'id': instance.id,
                'bytes_sent': int(bytes_sent),
                'bytes_recv': int(bytes_recv)
            }
        })

    def get_pid(self):
        """PIDs"""
        Pids.objects.all().delete()  # Clear all previous records
        pid_add_list = []
        data_list = []
        for proc in psutil.process_iter(['pid', 'name', 'memory_info', 'create_time', 'memory_percent', 'cpu_percent']):
            try:
                info = proc.info
                pid = info['pid']
                if pid == 0:
                    continue
                name = info['name']
                mem_rss = info['memory_info'].rss
                create_time = info['create_time']
                mem_percent = info['memory_percent']
                cpu_percent = info['cpu_percent']
                item = {
                    'pid': int(pid),
                    'name': str(name),
                    'memory': int(mem_rss),
                    'create_time': datetime.fromtimestamp(create_time),
                    'memory_usage': round(float(mem_percent), 2),
                    'cpu_usage': round(float(cpu_percent), 2)
                }
                pid_add = Pids(**item)
                pid_add_list.append(pid_add)
                data_list.append(item)
                bomiot_signals.send(sender=Pids, msg={
                                                    'models': 'Pids',
                                                    'type': 'created',
                                                    'data': item
                                                })
            except (psutil.NoSuchProcess, psutil.AccessDenied, psutil.ZombieProcess):
                continue
        Pids.objects.bulk_create(pid_add_list, batch_size=200)  # Bulk create PIDs to improve performance


    def cteate_pypi_database(self, data):
        return PyPi(
            date=data.get('date'),
            downloads=data.get('downloads'),
            percent=data.get('percent'),
            category=data.get('category')
        )

    def res_pypi_data(self, data):
        data["date"] = datetime.strptime(data.get('date'), '%Y-%m-%d')
        data["percent"] = float(data.get('percent').split('%')[0])
        return data

    def get_pypi_stats(self):
        today = datetime.now().date()
        pypi_today_qs = PyPi.objects.filter(created_time__date=today)
        if pypi_today_qs.exists() is False:
            PyPi.objects.exclude(created_time__date=today).delete()
            df = pypistats.overall("bomiot", total='daily', format="pandas")
            res = df.to_dict(orient="records")
            pypi_list = []
            for record in res:
                if record.get('category') == 'Total':
                    continue
                data = self.res_pypi_data(record)
                pypi_obj = self.cteate_pypi_database(data)
                pypi_list.append(pypi_obj)
            pypi_list.sort(key=lambda x: x.date, reverse=True)
            PyPi.objects.bulk_create(pypi_list, batch_size=200)

    def monitor_server(self):
        """Monitor server status"""
        while True:
            sleep(1)
            self.get_cpu_info()
            sleep(1)
            self.get_memory_info()
            sleep(1)
            self.get_disk_info()
            sleep(1)
            self.get_network_info()
            sleep(1)
            self.get_pid()
            sleep(1)
            self.get_pypi_stats()
            sleep(60)  # Perform monitoring every 60 seconds


def start_monitoring():
    """Start the server monitoring thread"""
    server_manager = ServerManager()
    monitoring_thread = threading.Thread(target=server_manager.monitor_server, daemon=True)
    monitoring_thread.start()



from django.contrib.auth import get_user_model
from django.db.models import JSONField
from django_filters import CharFilter, NumberFilter, DateFilter, DateTimeFilter, BooleanFilter, RangeFilter
from django_filters import FilterSet
from bomiot.server.core import models

# Get user model
User = get_user_model()

# define JSONField filter support more lookup_expr
def generate_jsonfield_filter(lookup_expr, filter_class):
    """
    create JSONField filter parameter
    :param lookup_expr: get（exact, icontains...）
    :return: filter parameter
    """
    return {
        JSONField: {
            'filter_class': filter_class,
            'extra': lambda f: {
                'lookup_expr': lookup_expr
            },
        }
    }

JSONFIELD_FILTER_OVERRIDE = {
    # String filters
    **generate_jsonfield_filter('exact', CharFilter),
    **generate_jsonfield_filter('iexact', CharFilter),
    **generate_jsonfield_filter('contains', CharFilter),
    **generate_jsonfield_filter('icontains', CharFilter),
    # Number filters (support integers and floats)
    **generate_jsonfield_filter('exact', NumberFilter),  # exact match for numbers (int/float)
    **generate_jsonfield_filter('lt', NumberFilter),  # less than
    **generate_jsonfield_filter('lte', NumberFilter),  # less than or equal
    **generate_jsonfield_filter('gt', NumberFilter),  # greater than
    **generate_jsonfield_filter('gte', NumberFilter),  # greater than or equal
    # Boolean filters
    **generate_jsonfield_filter('exact', BooleanFilter),  # exact match for boolean values
    # Date filters
    **generate_jsonfield_filter('exact', DateFilter),  # exact match for dates
    **generate_jsonfield_filter('lt', DateFilter),  # less than
    **generate_jsonfield_filter('lte', DateFilter),  # less than or equal
    **generate_jsonfield_filter('gt', DateFilter),  # greater than
    **generate_jsonfield_filter('gte', DateFilter),  # greater than or equal
    # DateTime filters
    **generate_jsonfield_filter('exact', DateTimeFilter),  # exact match for datetime
    **generate_jsonfield_filter('lt', DateTimeFilter),  # less than
    **generate_jsonfield_filter('lte', DateTimeFilter),  # less than or equal
    **generate_jsonfield_filter('gt', DateTimeFilter),  # greater than
    **generate_jsonfield_filter('gte', DateTimeFilter),  # greater than or equal
    # Range filters
    **generate_jsonfield_filter('range', RangeFilter),  # range filter for numbers, dates, etc.
}

class UserFilter(FilterSet):
    """
    User filter
    """
    class Meta:
        model = models.User
        fields = '__all__'
        filter_overrides = JSONFIELD_FILTER_OVERRIDE


class PermissionFilter(FilterSet):
    """
    Permission filter
    """
    class Meta:
        model = models.Permission
        fields = '__all__'
        filter_overrides = JSONFIELD_FILTER_OVERRIDE


class FileFilter(FilterSet):
    """
    File filter
    """
    class Meta:
        model = models.Files
        fields = '__all__'
        filter_overrides = JSONFIELD_FILTER_OVERRIDE


class TeamFilter(FilterSet):
    """
    Team filter
    """
    class Meta:
        model = models.Team
        fields = '__all__'
        filter_overrides = JSONFIELD_FILTER_OVERRIDE


class DepartmentFilter(FilterSet):
    """
    Team filter
    """
    class Meta:
        model = models.Department
        fields = '__all__'
        filter_overrides = JSONFIELD_FILTER_OVERRIDE


class APIFilter(FilterSet):
    """
    API filter
    """
    class Meta:
        model = models.API
        fields = '__all__'
        filter_overrides = JSONFIELD_FILTER_OVERRIDE


class ExampleFilter(FilterSet):
    """
    Example filter
    """
    class Meta:
        model = models.Example
        fields = '__all__'
        filter_overrides = JSONFIELD_FILTER_OVERRIDE


class PidsFilter(FilterSet):
    """
    Pids filter
    """
    class Meta:
        model = models.Pids
        fields = '__all__'
        filter_overrides = JSONFIELD_FILTER_OVERRIDE


class PyPiFilter(FilterSet):
    """
    Pids filter
    """
    class Meta:
        model = models.PyPi
        fields = '__all__'
        filter_overrides = JSONFIELD_FILTER_OVERRIDE


class CPUFilter(FilterSet):
    """
    CPU filter
    """
    class Meta:
        model = models.CPU
        fields = '__all__'
        filter_overrides = JSONFIELD_FILTER_OVERRIDE


class MemoryFilter(FilterSet):
    """
    Memory filter
    """
    class Meta:
        model = models.Memory
        fields = '__all__'
        filter_overrides = JSONFIELD_FILTER_OVERRIDE


class DiskFilter(FilterSet):
    """
    Disk filter
    """
    class Meta:
        model = models.Disk
        fields = '__all__'
        filter_overrides = JSONFIELD_FILTER_OVERRIDE



class NetworkFilter(FilterSet):
    """
    Network filter
    """
    class Meta:
        model = models.Network
        fields = '__all__'
        filter_overrides = JSONFIELD_FILTER_OVERRIDE


class GoodsFilter(FilterSet):
    """
    Goods filter
    """
    class Meta:
        model = models.Goods
        fields = '__all__'
        filter_overrides = JSONFIELD_FILTER_OVERRIDE


class BinFilter(FilterSet):
    """
    Bin filter
    """
    class Meta:
        model = models.Bin
        fields = '__all__'
        filter_overrides = JSONFIELD_FILTER_OVERRIDE


class StockFilter(FilterSet):
    """
    Stock filter
    """
    class Meta:
        model = models.Stock
        fields = '__all__'
        filter_overrides = JSONFIELD_FILTER_OVERRIDE


class CapitalFilter(FilterSet):
    """
    Capital filter
    """
    class Meta:
        model = models.Capital
        fields = '__all__'
        filter_overrides = JSONFIELD_FILTER_OVERRIDE


class SupplierFilter(FilterSet):
    """
    Supplier filter
    """
    class Meta:
        model = models.Supplier
        fields = '__all__'
        filter_overrides = JSONFIELD_FILTER_OVERRIDE


class CustomerFilter(FilterSet):
    """
    Customer filter
    """
    class Meta:
        model = models.Customer
        fields = '__all__'
        filter_overrides = JSONFIELD_FILTER_OVERRIDE


class ASNFilter(FilterSet):
    """
    ASN filter
    """
    class Meta:
        model = models.ASN
        fields = '__all__'
        filter_overrides = JSONFIELD_FILTER_OVERRIDE


class DNFilter(FilterSet):
    """
    DN filter
    """
    class Meta:
        model = models.DN
        fields = '__all__'
        filter_overrides = JSONFIELD_FILTER_OVERRIDE


class PurchaseFilter(FilterSet):
    """
    Purchase filter
    """
    class Meta:
        model = models.Purchase
        fields = '__all__'
        filter_overrides = JSONFIELD_FILTER_OVERRIDE


class BarFilter(FilterSet):
    """
    Bar filter
    """
    class Meta:
        model = models.Bar
        fields = '__all__'
        filter_overrides = JSONFIELD_FILTER_OVERRIDE


class FeeFilter(FilterSet):
    """
    Fee filter
    """
    class Meta:
        model = models.Fee
        fields = '__all__'
        filter_overrides = JSONFIELD_FILTER_OVERRIDE


class DriverFilter(FilterSet):
    """
    Driver filter
    """
    class Meta:
        model = models.Driver
        fields = '__all__'
        filter_overrides = JSONFIELD_FILTER_OVERRIDE


import importlib.metadata
import importlib.util

from os import listdir
from os.path import join, isdir, exists, isfile
from django.contrib import admin
from django.urls import path, include, re_path
from django.views.generic.base import TemplateView
from django.contrib.staticfiles.views import serve
from django.views.static import serve as static_serve
from django.conf import settings
from bomiot.server.server import views
from bomiot.server.server.pkgcheck import pkg_check, cwd_check, ignore_pkg, ignore_cwd
from configparser import ConfigParser
from pathlib import Path
from django.urls import resolve, Resolver404


def url_exists(url_data):
    try:
        resolve(url_data)
        return True
    except Resolver404:
        return False

def return_static(request, path, insecure=True, **kwargs):
    return serve(request, path, insecure, **kwargs)

templates_config = ConfigParser()
templates_config.read(join(settings.WORKING_SPACE, 'setup.ini'), encoding='utf-8')
templates_dir_name = templates_config.get('templates', 'name', fallback='templates/dist/spa/index.html')

urlpatterns = [
    # path('admin/', admin.site.urls),
    path('', TemplateView.as_view(template_name=templates_dir_name)),
    path('test/', views.test, name='test'),
    path('login/', views.logins, name='login'),
    path('logout/', views.logouts, name='logout'),
    path('checktoken/', views.check_token, name='check_token'),
    path('md/<str:mddocs>', views.mdurl, name='markdown'),
    path('core/', include('bomiot.server.core.urls')),
]

urlpatterns += [
    path('favicon.ico', views.favicon, name='favicon'),
    re_path('^css/.*$', views.statics, name='css'),
    re_path('^js/.*$', views.statics, name='js'),
    re_path('^assets/.*$', views.statics, name='assets'),
    re_path('^statics/.*$', views.statics, name='statics'),
    re_path('^fonts/.*$', views.statics, name='fonts'),
    re_path('^icons/.*$', views.statics, name='icons'),
    re_path(r'^static/(?P<path>.*)$', return_static, name='static'),
    re_path(r'^media/(?P<path>.*)$', static_serve, {'document_root': settings.MEDIA_ROOT}),
]

urlpatterns += [
    path('.well-known/appspecific/com.chrome.devtools.json', views.google),
]

all_packages = [dist.metadata['Name'] for dist in importlib.metadata.distributions()]
res_pkg_list = list(set([name.lower() for name in all_packages]).difference(set(ignore_pkg())))
pkg_squared = list(map(lambda data: pkg_check(data), res_pkg_list))
filtered_pkg_squared = list(filter(lambda x: x is not None, pkg_squared))

current_path = list(set([p for p in listdir(settings.WORKING_SPACE) if isdir(p)]).difference(set(ignore_cwd())))
cur_squared = list(map(lambda data: cwd_check(data), current_path))
filtered_current_path = list(filter(lambda y: y is not None, cur_squared))

if len(filtered_pkg_squared) > 0:
    for module in filtered_pkg_squared:
        try:
            spec = importlib.util.find_spec(f'{module}.urls')
            if spec is None:
                continue
            urls_module = importlib.import_module(f'{module}.urls')
            if not hasattr(urls_module, 'urlpatterns'):
                continue
            urlpatterns += [
                path(f'{module}/', include(f'{module}.urls'))
            ]
        except Exception as e:
            continue

if len(filtered_current_path) > 0:
    for module_name in filtered_current_path:
        app_mode_config = ConfigParser()
        app_mode_config.read(join(settings.WORKING_SPACE, module_name, 'bomiotconf.ini'), encoding='utf-8')
        app_mode = app_mode_config.get('mode', 'name')
        if app_mode == 'plugins':
            try:
                spec = importlib.util.find_spec(f'{module_name}.urls')
                if spec is None:
                    continue
                urls_module = importlib.import_module(f'{module_name}.urls')
                if not hasattr(urls_module, 'urlpatterns'):
                    continue
                urlpatterns += [
                    path(f'{module_name}/', include(f'{module_name}.urls'))
                ]
            except Exception as e:
                continue
        elif app_mode == 'project':
            if module_name == settings.PROJECT_NAME:
                project_path = join(settings.WORKING_SPACE, settings.PROJECT_NAME)
                root_path = Path(project_path)
                exclude_dirs = {
                    '__pycache__', 'static', 'media', 'templates', 'language',
                    'migrations', 'tests', 'test', 'docs', 'documentation'
                }
                url_include_list = [p for p in root_path.iterdir() if p.is_dir() and p.name not in exclude_dirs]
                for url in url_include_list:
                    app_name = url.name
                    try:
                        include_path = f'{module_name}.{app_name}.urls'
                        spec = importlib.util.find_spec(include_path)
                        if spec is None:
                            continue
                        urls_module = importlib.import_module(include_path)
                        if not hasattr(urls_module, 'urlpatterns'):
                            continue
                        url_pattern = f'{app_name}/'
                        if any(pattern.pattern.regex.pattern.startswith(url_pattern) for pattern in urlpatterns):
                            continue
                        urlpatterns.append(
                            path(url_pattern, include(include_path))
                        )
                    except Exception as e:
                        import traceback
                        traceback.print_exc()
                        continue


import jwt
import datetime
from django.conf import settings

# from settings import JWT_SALT 
JWT_SALT = getattr(settings, "JWT_SALT", settings.SECRET_KEY)


def create_token(payload):
    """
    create JWT Token
    :param payload: JWT Token data
    :return: JWT Token
    """
    headers = {
        "type": "JWT",
        "alg": "HS256"
    }
    # set Token outtime
    payload['exp'] = datetime.datetime.now() + datetime.timedelta(seconds=settings.USER_JWT_TIME)
    # create Token
    token = jwt.encode(payload=payload, key=JWT_SALT, algorithm="HS256", headers=headers)
    return token


def parse_payload(token):
    """
    parse JWT Token
    :param token: JWT Token
    :return: include status, data, error
    """
    result = {"status": False, "data": None, "error": None}
    try:
        # verify Token and decode payload
        verified_payload = jwt.decode(token, JWT_SALT, algorithms="HS256", options={"verify_exp": True})
        result["status"] = True
        result['data'] = verified_payload
    except jwt.ExpiredSignatureError:
        result['detail'] = 'Token Expired'
    except jwt.DecodeError:
        result['detail'] = 'Token Authentication Failed'
    except jwt.InvalidTokenError:
        result['detail'] = 'Illegal Token'
    except Exception as e:
        result['detail'] = f"Unknown error: {str(e)}"
    return result


import os
import django
from django.core.management import call_command


def migrate():
    """
    Execute database migrations
    """
    os.environ.setdefault("DJANGO_SETTINGS_MODULE", "bomiot.server.server.settings")
    os.environ.setdefault('RUN_MAIN', 'true')
    django.setup()
    
    try:
        call_command('migrate')
    except Exception as e:
        print(f"Error during migration: {e}") 


from django.apps import AppConfig
from django.db import connections, connection, transaction
from django.db.migrations.executor import MigrationExecutor
from django.db.models.signals import post_migrate
import os
from os.path import join
from time import sleep
import threading

class CoreConfig(AppConfig):
    """
    Core application configuration for the bomiot server.
    """
    name = 'bomiot.server.core'

    def ready(self):
        from django.conf import settings
        from bomiot.server.core.models import API
        from bomiot.server.core import signal
        post_migrate.connect(do_init_data, sender=self)
        try:
            if not API.objects.filter().exists():
                init_api()
        except Exception as e:
            print(f"Initial API initialization failed: {e}")
        workers = int(os.environ.get('WORKERS', 0))
        if workers > 0:
            lockfile = f"{join(settings.WORKING_SPACE, 'bomiot_ready.lock')}"
            try:
                fd = os.open(lockfile, os.O_CREAT | os.O_EXCL | os.O_RDWR)
                from bomiot.server.server.views import init_permission
                from bomiot.server.core.scheduler import sm
                from bomiot.server.core.observer import ob
                from bomiot.server.core.server_monitor import start_monitoring
                from bomiot.server.server.views import init_permission
                from bomiot.server.core.signal import bomiot_signals
                start_monitoring()
                sm.start()
                ob.start()
                def backgrun_init():
                    init_permission()
                init_thread = threading.Thread(target=backgrun_init, daemon=True)
                init_thread.start()

                print('')
                print("  $$$$$$    $$$$$   $$$       $$$  $$   $$$$$   $$$$$$")
                print("  $$   $$  $$   $$  $$ $     $ $$  $$  $$   $$    $$")
                print("  $$$$$$$  $$   $$  $$  $   $  $$  $$  $$   $$    $$")
                print("  $$   $$  $$   $$  $$   $ $   $$  $$  $$   $$    $$")
                print("  $$$$$$    $$$$$   $$    $    $$  $$   $$$$$     $$")
                print('')

            except FileExistsError:
                pass
                
def do_init_data(sender, **kwargs):
    init_api()


def init_api():
    try:
        from bomiot.server.core.models import API
        if API.objects.filter().exists():
            if API.objects.filter().count() != 52:
                API.objects.all().delete()
                create_api_data()
        else:
            create_api_data()
    except Exception as e:
        print(f"Init database error: {e}")


def create_api_data():
    from bomiot.server.core.models import API
    init_data = [
        API(id=1, method='GET', api='/core/example/', func_name='example_get', name="Get Example List"),
        API(id=2, method='POST', api='/core/example/create/', func_name='example_create', name="Create Example"),
        API(id=3, method='POST', api='/core/example/update/', func_name='example_update', name="Update Example"),
        API(id=4, method='POST', api='/core/example/delete/', func_name='example_delete', name="Delete Example"),
        API(id=5, method='GET', api='/core/goods/', func_name='goods_get', name="Get Goods List"),
        API(id=6, method='POST', api='/core/goods/create/', func_name='goods_create', name="Create Goods"),
        API(id=7, method='POST', api='/core/goods/update/', func_name='goods_update', name="Update Goods"),
        API(id=8, method='POST', api='/core/goods/delete/', func_name='goods_delete', name="Delete Goods"),
        API(id=9, method='GET', api='/core/bin/', func_name='bin_get', name="Get Bin List"),
        API(id=10, method='POST', api='/core/bin/create/', func_name='bin_create', name="Create Bin"),
        API(id=11, method='POST', api='/core/bin/update/', func_name='bin_update', name="Update Bin"),
        API(id=12, method='POST', api='/core/bin/delete/', func_name='bin_delete', name="Delete Bin"),
        API(id=13, method='GET', api='/core/stock/', func_name='stock_get', name="Get Stock List"),
        API(id=14, method='POST', api='/core/stock/create/', func_name='stock_create', name="Create Stock"),
        API(id=15, method='POST', api='/core/stock/update/', func_name='stock_update', name="Update Stock"),
        API(id=16, method='POST', api='/core/stock/delete/', func_name='stock_delete', name="Delete Stock"),
        API(id=17, method='GET', api='/core/capital/', func_name='capital_get', name="Get Capital List"),
        API(id=18, method='POST', api='/core/capital/create/', func_name='capital_create', name="Create Capital"),
        API(id=19, method='POST', api='/core/capital/update/', func_name='capital_update', name="Update Capital"),
        API(id=20, method='POST', api='/core/capital/delete/', func_name='capital_delete', name="Delete Capital"),
        API(id=21, method='GET', api='/core/supplier/', func_name='supplier_get', name="Get Supplier List"),
        API(id=22, method='POST', api='/core/supplier/create/', func_name='supplier_create', name="Create Supplier"),
        API(id=23, method='POST', api='/core/supplier/update/', func_name='supplier_update', name="Update Supplier"),
        API(id=24, method='POST', api='/core/supplier/delete/', func_name='supplier_delete', name="Delete Supplier"),
        API(id=25, method='GET', api='/core/customer/', func_name='customer_get', name="Get Customer List"),
        API(id=26, method='POST', api='/core/customer/create/', func_name='customer_create', name="Create Customer"),
        API(id=27, method='POST', api='/core/customer/update/', func_name='customer_update', name="Update Customer"),
        API(id=28, method='POST', api='/core/customer/delete/', func_name='customer_delete', name="Delete Customer"),
        API(id=29, method='GET', api='/core/asn/', func_name='asn_get', name="Get ASN List"),
        API(id=30, method='POST', api='/core/asn/create/', func_name='asn_create', name="Create ASN"),
        API(id=31, method='POST', api='/core/asn/update/', func_name='asn_update', name="Update ASN"),
        API(id=32, method='POST', api='/core/asn/delete/', func_name='asn_delete', name="Delete ASN"),
        API(id=33, method='GET', api='/core/dn/', func_name='dn_get', name="Get DN List"),
        API(id=34, method='POST', api='/core/dn/create/', func_name='dn_create', name="Create DN"),
        API(id=35, method='POST', api='/core/dn/update/', func_name='dn_update', name="Update DN"),
        API(id=36, method='POST', api='/core/dn/delete/', func_name='dn_delete', name="Delete DN"),
        API(id=37, method='GET', api='/core/purchase/', func_name='purchase_get', name="Get Purchase List"),
        API(id=38, method='POST', api='/core/purchase/create/', func_name='purchase_create', name="Create Purchase"),
        API(id=39, method='POST', api='/core/purchase/update/', func_name='purchase_update', name="Update Purchase"),
        API(id=40, method='POST', api='/core/purchase/delete/', func_name='purchase_delete', name="Delete Purchase"),
        API(id=41, method='GET', api='/core/bar/', func_name='bar_get', name="Get Bar List"),
        API(id=42, method='POST', api='/core/bar/create/', func_name='bar_create', name="Create Bar"),
        API(id=43, method='POST', api='/core/bar/update/', func_name='bar_update', name="Update Bar"),
        API(id=44, method='POST', api='/core/bar/delete/', func_name='bar_delete', name="Delete Bar"),
        API(id=45, method='GET', api='/core/fee/', func_name='fee_get', name="Get Fee List"),
        API(id=46, method='POST', api='/core/fee/create/', func_name='fee_create', name="Create Fee"),
        API(id=47, method='POST', api='/core/fee/update/', func_name='fee_update', name="Update Fee"),
        API(id=48, method='POST', api='/core/fee/delete/', func_name='fee_delete', name="Delete Fee"),
        API(id=49, method='GET', api='/core/driver/', func_name='driver_get', name="Get Driver List"),
        API(id=50, method='POST', api='/core/driver/create/', func_name='driver_create', name="Create Driver"),
        API(id=51, method='POST', api='/core/driver/update/', func_name='driver_update', name="Update Driver"),
        API(id=52, method='POST', api='/core/driver/delete/', func_name='driver_delete', name="Delete Driver"),
    ]
    API.objects.bulk_create(init_data, batch_size=200)


import os
from django.urls import path
from bomiot.server.core import client, views, handler
from bomiot.server.function import goods
from bomiot.server.function import bin
from bomiot.server.function import stock
from bomiot.server.function import capital
from bomiot.server.function import supplier
from bomiot.server.function import customer
from bomiot.server.function import asn
from bomiot.server.function import dn
from bomiot.server.function import purchase
from bomiot.server.function import bar
from bomiot.server.function import fee
from bomiot.server.function import driver


urlpatterns = [
    path(r'user/', views.UserList.as_view({"get": "list"}), name="Get User List"),
    path(r'user/permission/', views.PermissionList.as_view({"get": "list"})),
    path(r'user/create/', views.UserCreate.as_view({"post": "create"}), name="Create One User"),
    path(r'user/changepwd/', views.UserChangePWD.as_view({"post": "create"}), name="Change Password"),
    path(r'user/team/', views.UserSetTeam.as_view({"post": "create"}), name="Set Team For User"),
    path(r'user/department/', views.UserSetDepartment.as_view({"post": "create"}), name="Set Department For User"),
    path(r'user/lock/', views.UserLock.as_view({"post": "create"}), name="Lock & Unlock User"),
    path(r'user/delete/', views.UserDelete.as_view({"post": "create"}), name="Delete One User"),
    path(r'user/upload/', views.UserUpload.as_view({"post": "create"})),
    path(r'user/files/', views.UserFiles.as_view({"get": "list"})),
    path(r'user/files/share/', views.FileShare.as_view({"post": "create"})),
    path(r'user/files/delete/', views.DeleteFile.as_view({"post": "create"}))
]

urlpatterns += [
    path(r'team/', views.TeamList.as_view({"get": "list"}), name="Get Team List"),
    path(r'team/create/', views.TeamCreate.as_view({"post": "create"}), name="Create One Team"),
    path(r'team/setpermission/', views.TeamPermission.as_view({"post": "create"}), name="Set Permission For Team"),
    path(r'team/change/', views.TeamChange.as_view({"post": "create"}), name="Change Team"),
    path(r'team/delete/', views.TeamDelete.as_view({"post": "create"}), name="Delete Team")
]

urlpatterns += [
    path(r'department/', views.DepartmentList.as_view({"get": "list"}), name="Get Department List"),
    path(r'department/create/', views.DepartmentCreate.as_view({"post": "create"}), name="Create Department"),
    path(r'department/change/', views.DepartmentChange.as_view({"post": "create"}), name="Change Department"),
    path(r'department/delete/', views.DepartmentDelete.as_view({"post": "create"}), name="Delete Department")
]

urlpatterns += [
    path(r'pypi/', views.PyPiList.as_view({"get": "list"})),
    path(r'pypi/charts/', views.PyPiCharts.as_view({"get": "list"})),
]

urlpatterns += [
    path(r'pid/', client.PIDList.as_view({"get": "list"})),
    path(r'cpu/', client.CPUList.as_view({"get": "list"})),
    path(r'memory/', client.MemoryList.as_view({"get": "list"})),
    path(r'disk/', client.DiskList.as_view({"get": "list"})),
    path(r'network/', client.NetworkList.as_view({"get": "list"})),
    path(r'server/echarts/', client.ServerCharts.as_view({"get": "list"})),
    path(r'server/pidcharts/', client.PIDCharts.as_view({"get": "list"})),
]

urlpatterns += [
    path(r'example/', handler.ExampleList.as_view({"get": "list"})),
    path(r'example/create/', handler.ExampleCreate.as_view({"post": "create"})),
    path(r'example/update/', handler.ExampleUpdate.as_view({"post": "create"})),
    path(r'example/delete/', handler.ExampleDelete.as_view({"post": "create"}))
]

urlpatterns += [
    path(r'goods/', goods.GoodsList.as_view({"get": "list"}), name="Get Goods List"),
    path(r'goods/create/', goods.GoodsCreate.as_view({"post": "create"}), name="Create Goods"),
    path(r'goods/update/', goods.GoodsUpdate.as_view({"post": "create"}), name="Update Goods"),
    path(r'goods/delete/', goods.GoodsDelete.as_view({"post": "create"}), name="Delete Goods")
]

urlpatterns += [
    path(r'bin/', bin.BinList.as_view({"get": "list"}), name="Get Bin List"),
    path(r'bin/create/', bin.BinCreate.as_view({"post": "create"}), name="Create Bin"),
    path(r'bin/update/', bin.BinUpdate.as_view({"post": "create"}), name="Update Bin"),
    path(r'bin/delete/', bin.BinDelete.as_view({"post": "create"}), name="Delete Bin")
]

urlpatterns += [
    path(r'stock/', stock.StockList.as_view({"get": "list"}), name="Get Stock List"),
    path(r'stock/create/', stock.StockCreate.as_view({"post": "create"}), name="Create Stock"),
    path(r'stock/update/', stock.StockUpdate.as_view({"post": "create"}), name="Update Stock"),
    path(r'stock/delete/', stock.StockDelete.as_view({"post": "create"}), name="Delete Stock")
]

urlpatterns += [
    path(r'capital/', capital.CapitalList.as_view({"get": "list"}), name="Get Capital List"),
    path(r'capital/create/', capital.CapitalCreate.as_view({"post": "create"}), name="Create Capital"),
    path(r'capital/update/', capital.CapitalUpdate.as_view({"post": "create"}), name="Update Capital"),
    path(r'capital/delete/', capital.CapitalDelete.as_view({"post": "create"}), name="Delete Capital")
]

urlpatterns += [
    path(r'supplier/', supplier.SupplierList.as_view({"get": "list"}), name="Get Supplier List"),
    path(r'supplier/create/', supplier.SupplierCreate.as_view({"post": "create"}), name="Create Supplier"),
    path(r'supplier/update/', supplier.SupplierUpdate.as_view({"post": "create"}), name="Update Supplier"),
    path(r'supplier/delete/', supplier.SupplierDelete.as_view({"post": "create"}), name="Delete Supplier")
]

urlpatterns += [
    path(r'customer/', customer.CustomerList.as_view({"get": "list"}), name="Get Customer List"),
    path(r'customer/create/', customer.CustomerCreate.as_view({"post": "create"}), name="Create Customer"),
    path(r'customer/update/', customer.CustomerUpdate.as_view({"post": "create"}), name="Update Customer"),
    path(r'customer/delete/', customer.CustomerDelete.as_view({"post": "create"}), name="Delete Customer")
]

urlpatterns += [
    path(r'asn/', asn.ASNList.as_view({"get": "list"}), name="Get ASN List"),
    path(r'asn/create/', asn.ASNCreate.as_view({"post": "create"}), name="Create ASN"),
    path(r'asn/update/', asn.ASNUpdate.as_view({"post": "create"}), name="Update ASN"),
    path(r'asn/delete/', asn.ASNDelete.as_view({"post": "create"}), name="Delete ASN")
]

urlpatterns += [
    path(r'dn/', dn.DNList.as_view({"get": "list"}), name="Get DN List"),
    path(r'dn/create/', dn.DNCreate.as_view({"post": "create"}), name="Create DN"),
    path(r'dn/update/', dn.DNUpdate.as_view({"post": "create"}), name="Update DN"),
    path(r'dn/delete/', dn.DNDelete.as_view({"post": "create"}), name="Delete DN")
]

urlpatterns += [
    path(r'purchase/', purchase.PurchaseList.as_view({"get": "list"}), name="Get Purchase List"),
    path(r'purchase/create/', purchase.PurchaseCreate.as_view({"post": "create"}), name="Create Purchase"),
    path(r'purchase/update/', purchase.PurchaseUpdate.as_view({"post": "create"}), name="Update Purchase"),
    path(r'purchase/delete/', purchase.PurchaseDelete.as_view({"post": "create"}), name="Delete Purchase")
]

urlpatterns += [
    path(r'bar/', bar.BarList.as_view({"get": "list"}), name="Get Bar List"),
    path(r'bar/create/', bar.BarCreate.as_view({"post": "create"}), name="Create Bar"),
    path(r'bar/update/', bar.BarUpdate.as_view({"post": "create"}), name="Update Bar"),
    path(r'bar/delete/', bar.BarDelete.as_view({"post": "create"}), name="Delete Bar")
]

urlpatterns += [
    path(r'fee/', fee.FeeList.as_view({"get": "list"}), name="Get Fee List"),
    path(r'fee/create/', fee.FeeCreate.as_view({"post": "create"}), name="Create Fee"),
    path(r'fee/update/', fee.FeeUpdate.as_view({"post": "create"}), name="Update Fee"),
    path(r'fee/delete/', fee.FeeDelete.as_view({"post": "create"}), name="Delete Fee")
]

urlpatterns += [
    path(r'driver/', driver.DriverList.as_view({"get": "list"}), name="Get Driver List"),
    path(r'driver/create/', driver.DriverCreate.as_view({"post": "create"}), name="Create Driver"),
    path(r'driver/update/', driver.DriverUpdate.as_view({"post": "create"}), name="Update Driver"),
    path(r'driver/delete/', driver.DriverDelete.as_view({"post": "create"}), name="Delete Driver")
]

urlpatterns += [
    path(r'api/', views.APIList.as_view({"get": "list"})),
    path(r'api/change/', views.APIChange.as_view({"post": "create"})),
]


from os.path import join, exists, isfile, isdir
from os import makedirs, getcwd, listdir
import os
import sys
import shutil
from pathlib import Path
from .init import create_file
from configparser import ConfigParser
from .copyfile import copy_files
from bomiot.server.server.pkgcheck import pkg_check, cwd_check, ignore_pkg, ignore_cwd
import importlib.util
import subprocess


def copy_project(folder: str):
    """
    marketplace
    :param folder:
    :return:
    """
    if len(sys.argv) < 3:
        print('Please enter your project name')
    else:
        project_path = join(getcwd(), sys.argv[2])
        if exists(project_path):
            print('Project directory already exists')
        else:
            try:
                module_path = importlib.util.find_spec(sys.argv[2]).origin
                list_module_path = Path(module_path).resolve().parent
                pkg_config_check = ConfigParser()
                pkg_config_check.read(join(list_module_path, 'bomiotconf.ini'), encoding='utf-8')
                app_mode = pkg_config_check.get('mode', 'name', fallback='plugins')
                if app_mode == 'project':
                    copy_files(list_module_path, join(getcwd(), sys.argv[2]))
                    setup_config = ConfigParser()
                    setup_config.read(join(join(getcwd()), 'setup.ini'), encoding='utf-8')
                    setup_config.set('project', 'name', folder)
                    setup_config.write(open(join(join(getcwd()), 'setup.ini'), "wt"))
                else:
                    print('No project found in the current environment, please check your environment or install the project first.')
            except subprocess.CalledProcessError as e:
                print(f"No package named {sys.argv[2]} found, please check your package name or install it first.")
                sys.exit(1)
            print(f'Initialized project workspace {sys.argv[2]}')


import json, orjson
import mimetypes
import os
import aiofiles
from django.conf import settings
from django.http import JsonResponse, HttpResponse, FileResponse, StreamingHttpResponse
from wsgiref.util import FileWrapper
from configparser import ConfigParser
from django.contrib.auth import authenticate, login, logout
from django.contrib.auth.decorators import login_required
from django.contrib.auth import get_user_model
from django.core.cache import cache
from bomiot.server.core.jwt_auth import create_token, parse_payload
from bomiot.server.core.models import Permission
from bomiot_message import login_message_return, others_message_return
from bomiot.server.server.pkgcheck import url_ignore
from os.path import join, isdir, exists
from os import listdir
import importlib.util
from pathlib import Path
from django.urls import get_resolver, URLPattern, URLResolver

User = get_user_model()

async def test(request):
    return JsonResponse({"msg": "This is Django API"})


def logins(request):
    data = json.loads(request.body.decode().replace("'", '"'))
    user_check = User.objects.filter(username=data.get('username'), is_delete=False)
    if user_check.exists() is False:
        return JsonResponse(login_message_return(request.META.get('HTTP_LANGUAGE', ''), "User not exists"))
    else:
        user = authenticate(username=data.get('username'), password=data.get('password'))
        context = {}
        if user:
            if user.is_active is True:
                login(request, user)
                user_info = {
                    "id": user.id,
                    "username": user.username,
                    "admin": user.is_superuser,
                    "permission": user.permission
                }
                token = create_token(user_info)
                context['token'] = token
                context['msg'] = others_message_return(request.META.get('HTTP_LANGUAGE', ''), 'Success Login')
                return JsonResponse(context)
            else:
                return JsonResponse(login_message_return(request.META.get('HTTP_LANGUAGE', ''), 'User is not active'))
        else:
            user_data = user_check.first()
            if user_data.request_limit < settings.REQUEST_LIMIT:
                user_data.request_limit += 1
                user_data.save()
                return JsonResponse(login_message_return(request.META.get('HTTP_LANGUAGE', ''), 'User or Password error'))
            else:
                user_data.is_active = False
                user_data.request_limit = 0
                user_data.save()
                return JsonResponse(login_message_return(request.META.get('HTTP_LANGUAGE', ''), 'User is not active'))


@login_required
async def logouts(request):
    if request.user.is_authenticated:
        logout(request)
        return JsonResponse({'msg': others_message_return(request.META.get('HTTP_LANGUAGE', ''), 'Welcome Back Again')})
    else:
        return JsonResponse(login_message_return(request.META.get('HTTP_LANGUAGE', ''), 'User Not Log In'))


async def check_token(request):
    token = request.META.get('HTTP_TOKEN')
    context = parse_payload(token)
    return JsonResponse(context)


async def mdurl(request, mddocs):
    language = request.META.get('HTTP_LANUAGE', '')
    if not mddocs.endswith('.md'):
        return JsonResponse({'detail': others_message_return(language, 'Only support markdown file')})
    folder_path = Path(settings.MEDIA_ROOT)
    all_files = [f.name for f in folder_path.iterdir() if f.is_file()]
    start_words = mddocs.split('.')
    md_check_list_all = []
    md_check_list_only = []
    for i in all_files:
        if i == mddocs:
            md_check_list_only.append(i)
        else:
            if i.startswith(start_words[0]) and i.endswith('.md'):
                md_check_list_all.append(i)
    if len(md_check_list_only) == 1:
        async def file_iterator():
            async with aiofiles.open(join(settings.MEDIA_ROOT, mddocs), 'rb') as f:
                while chunk := await f.read(8192):
                    yield chunk
        response = StreamingHttpResponse(file_iterator())
        response['Cache-Control'] = 'no-store, no-cache, must-revalidate, max-age=0'
        response['Pragma'] = 'no-cache'
        response['Expires'] = '0'
        return response
    else:
        if len(md_check_list_all) == 0:
            return JsonResponse({'detail': others_message_return(language, 'Markdown file not found')})
        async def file_iterator():
            async with aiofiles.open(join(settings.MEDIA_ROOT, mddocs), 'rb') as f:
                while chunk := await f.read(8192):
                    yield chunk
        response = StreamingHttpResponse(file_iterator())
        response['Cache-Control'] = 'no-store, no-cache, must-revalidate, max-age=0'
        response['Pragma'] = 'no-cache'
        response['Expires'] = '0'
        return response

def favicon(request):
    path = join(settings.MEDIA_ROOT, 'img', 'logo.png')
    resp = FileResponse(open(path, 'rb'))
    resp['Cache-Control'] = 'max-age=864000000000'
    return resp

def statics(request):
    if cache.has_key("templates_path") is False:
            CONFIG = ConfigParser()
            CONFIG.read(join(settings.WORKING_SPACE, 'setup.ini'), encoding='utf-8')
            templates_path = CONFIG.get('templates', 'name', fallback='templates/dist/spa/index.html')
            cache.set("templates_path", templates_path)
    else:
        templates_path = cache.get("templates_path")
    path = join(settings.WORKING_SPACE, settings.PROJECT_NAME)
    static_path_split = templates_path.split('/')[:-1]
    for i in static_path_split:
        if i == '':
            continue
        else:
            path = join(path, i)
    request_path = request.path_info.split('/')
    for j in request_path:
        if j == '':
            continue
        else:
            path = join(path, j)
    resp = FileResponse(open(path, 'rb'))
    resp['Cache-Control'] = 'max-age=864000000000'
    return resp

async def google(request):
    return JsonResponse({})

def queryset_to_json(queryset):
    data = list(queryset.values())
    return JsonResponse(data, safe=False)

def get_all_url(resolver=None, pre='/'):
    if resolver is None:
        resolver = get_resolver()
    for r in resolver.url_patterns:
        if isinstance(r, URLPattern):
            if '<pk>' in str(r.pattern):
                continue
            yield pre + str(r.pattern).replace('^', '').replace('$', ''), r.name
        if isinstance(r, URLResolver):
            yield from get_all_url(r, pre + str(r.pattern))

def permission_check(data, perm_obj_dict, api_list, user_objs):
    if str(data[0]) not in api_list and str(data[1]) != 'None':
        perm_obj, created = perm_obj_dict.get((str(data[0]), str(data[1])))
        if not perm_obj:
            perm_obj = Permission(api=str(data[0]), name=str(data[1]))
            perm_obj_dict[(str(data[0]), str(data[1]))] = perm_obj
        for user in user_objs:
            user.permission[str(data[1])] = str(data[0])
        return perm_obj

def init_permission():
    try:
        Permission.objects.all().delete()
        user_objs = list(User.objects.filter(is_superuser=True))
        media_root = settings.MEDIA_ROOT
        for user in user_objs:
            user_folder = join(media_root, user.username)
            if not exists(user_folder):
                os.makedirs(user_folder)
            user.permission = {}
        all_api_info = list(get_all_url())
        api_list = set(url_ignore())
        perm_obj_dict = {}
        perm_objs = []
        for data in all_api_info:
            if str(data[0]) not in api_list and str(data[1]) != 'None':
                perm_obj = Permission(api=str(data[0]), name=str(data[1]))
                perm_objs.append(perm_obj)
                for user in user_objs:
                    user.permission[str(data[1])] = str(data[0])
        if perm_objs:
            Permission.objects.bulk_create(perm_objs, batch_size=200)
        User.objects.bulk_update(user_objs, ['permission'], batch_size=100)
    except Exception as e:
        print(f"Error initializing permissions: {e}")


from flask import Flask, jsonify
from django.contrib.auth import get_user_model

User = get_user_model()

flask_app = Flask(__name__)

@flask_app.route("/test/")
def test():
    return jsonify({"msg": "This is a Flask API"})


import orjson
import ast
import aiofiles
import asyncio
import pandas as pd
from django.conf import settings
from os.path import join
import importlib.util
import importlib
from pathlib import Path
from bomiot_message import msg_message_return

def get_job_id(task):
    """
    construct job id
    :return: job id
    """
    return f'{task.job_id}'


def bytes2str(data):
    """
    bytes2str
    :param data: origin data
    :return: str
    """
    if isinstance(data, bytes):
        data = data.decode('utf-8')
    data = data.strip()
    return data


def load_dict(x, transformer=None):
    """
    convert to  dict
    :param x:
    :return:
    """
    if x is None or isinstance(x, dict):
        return x
    try:
        data = orjson.loads(x)
        if not transformer:
            def transformer(x): return x
        data = {k: transformer(v) for k, v in data.items()}
        return data
    except:
        return {}


def str2list(x, transformer=None):
    """
    convert to list
    :param x:
    :return:
    """
    if x is None or isinstance(x, list):
        return x
    try:
        data = orjson.loads(x)
        if not transformer:
            def transformer(x): return x
        data = list(map(lambda x: transformer(x), data))
        return data
    except:
        return []


def str2bool(v):
    """
    convert string to bool
    :param v:
    :return:
    """
    if isinstance(v, bool):
        return v
    if v.lower() in ('yes', 'true', 't', 'y', '1'):
        return True
    elif v.lower() in ('no', 'false', 'f', 'n', '0'):
        return False
    return True


def str2json(v):
    """
    convert str to json data
    :param v:
    :return:
    """
    try:
        return orjson.loads(v)
    except:
        return None


def str2dict(v):
    """
    convert str to dict data
    :param v:
    :return:
    """
    try:
        return orjson.loads(v)
    except:
        return {}


def str2body(v):
    """
    convert str to json data or keep original string
    :param v:
    :return:
    """
    try:
        return orjson.loads(v)
    except:
        return v


def str2str(v):
    """
    convert str to str, process for 'None', 'null', '',
    :param v:
    :return:
    """
    if v.lower() in ('none', 'null', 'undefined', 'nil', 'false'):
        return None
    return str(v)


def is_dict_empty(d):
    return not d.items()


def excel_to_json(path: str, skip=0, read_rows=31) -> list:
    df = pd.read_excel(path, nrows=0, header=0, engine='openpyxl')
    column_names = df.columns.tolist()
    df = pd.read_excel(path, skiprows=skip, nrows=read_rows, engine='openpyxl', names=column_names)
    data = df.to_json(orient="records", date_format="iso", date_unit='s', force_ascii=False)
    return orjson.loads(data)


def read_excel_title(path: str) -> list:
    df = pd.read_excel(path, nrows=0, header=0, engine='openpyxl')
    return df.columns.tolist()


def contains_value(data: dict, value) -> bool:
    return value in data.values()


def readable_file_size(size_in_bytes) -> str:
    for unit in ['B', 'KB', 'MB', 'GB', 'TB', 'PB']:
        if size_in_bytes < 1024.0:
            return f"{size_in_bytes:.2f} {unit}"
        size_in_bytes /= 1024.0
    return f"{size_in_bytes:.2f} EB"


def compare_dicts(dict1, dict2) -> dict:
    return {k: (dict1[k], dict2[k]) for k in dict1 if dict1[k] != dict2[k]}


def queryset_to_dict(queryset) -> list:
    """
    Convert a Django queryset to a list of dictionaries.
    :param queryset: Django queryset
    :return: List of dictionaries
    """
    return [{
        **flatten_json(obj),
        'created_time': obj['created_time'].strftime('%Y-%m-%d %H:%M:%S'),
        'updated_time': obj['updated_time'].strftime('%Y-%m-%d %H:%M:%S')
    } for obj in queryset.values('id', 'created_time', 'updated_time', 'data')]


def find_value_in_json(json_data, key_to_find):
    if isinstance(json_data, dict):
        for key, value in json_data.items():
            if key == key_to_find:
                return value
            if isinstance(value, (dict, list)):
                result = find_value_in_json(value, key_to_find)
                if result is not None:
                    return result
    elif isinstance(json_data, list):
        for item in json_data:
            result = find_value_in_json(item, key_to_find)
            if result is not None:
                return result
    return None


def flatten_json(nested_dict):
    return {
        **nested_dict['data'],
        'id': nested_dict['id'],
        'created_time': nested_dict['created_time'],
        'updated_time': nested_dict['updated_time']
    }


def is_empty_value(val):
    if isinstance(val, dict):
        return all(is_empty_value(v) for v in val.values())
    elif isinstance(val, list):
        return all(is_empty_value(v) for v in val)
    return val in [None, '', 'null', 'undefined', 'nil', [], {}, set()]


def all_fields_empty(json_dict):
    return is_empty_value(json_dict)


def check_method_in_file_by_ast(file_path, method_name):
    try:
        detail = {}
        with open(file_path, 'r', encoding='utf-8') as f:
            tree = ast.parse(f.read(), filename=file_path)
        found = False
        for node in ast.walk(tree):
            if isinstance(node, (ast.FunctionDef, ast.AsyncFunctionDef)):
                if node.name == method_name and isinstance(tree, ast.Module):
                    detail = {
                                "class": node.name,
                                "method": method_name,
                            }
                    found = True
                    break
            elif isinstance(node, ast.ClassDef):
                for item in node.body:
                    if isinstance(item, (ast.FunctionDef, ast.AsyncFunctionDef)):
                        if item.name == method_name:
                            detail = {
                                "class": node.name,
                                "method": method_name,
                            }
                            found = True
                            break
                if found:
                    break
        return found, detail

    except FileNotFoundError:
        print(f"'{file_path}' can not find")
        return False, detail
    except SyntaxError as e:
        print(f"'{file_path}' {e}")
        return False, detail
    except Exception as e:
        print(f"{e}")
        return False, detail

def receiver_callback(data, method) -> dict:
    if settings.PROJECT_NAME == 'bomiot':
        receiver_path = join(settings.BASE_DIR, 'core', 'receiver.py')
    else:
        receiver_path = join(settings.WORKING_SPACE, settings.PROJECT_NAME, 'receiver.py')
    receiver_check = check_method_in_file_by_ast(receiver_path, method)
    if receiver_check[0] is True:
        spec = importlib.util.spec_from_file_location("receiver", receiver_path)
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
        if 'class' in receiver_check[1]:
            try:
                target_class = getattr(module, receiver_check[1]['class'])
                instance = target_class()
                if hasattr(instance, receiver_check[1]['method']):
                    result = getattr(instance, receiver_check[1]['method'])(data)
                    return result
                else:
                    print(f"{type(receiver_check[1]).get('class')} class don't have {receiver_check[1]['method']} method")
            except AttributeError:
                print(f"class {type(receiver_check[1]).get('class')} can not find {receiver_path}")
        else:
            if hasattr(module, receiver_check[1]['method']):
                result = getattr(module, receiver_check[1]['method'])(data)
                return result
            else:
                print(f"{receiver_path} don't have {receiver_check[1]['method']} method")
    else:
        mode = data.get('mode')
        language = data.get('request').META.get('HTTP_LANGUAGE', 'en-US')
        if mode == 'get':
            return [
            ('results', data.get('data')),
        ]
        elif mode == 'create':
            return msg_message_return(language, "Success Create")
        elif mode == 'update':
            return msg_message_return(language, "Success Update")
        elif mode == 'delete':
            return msg_message_return(language, "Success Delete")
        

def receiver_file_callback(data, method) -> dict:
    if settings.PROJECT_NAME == 'bomiot':
        receiver_path = join(settings.BASE_DIR, 'core', 'files.py')
    else:
        receiver_path = join(settings.WORKING_SPACE, settings.PROJECT_NAME, 'files.py')
    receiver_check = check_method_in_file_by_ast(receiver_path, method)
    if receiver_check[0] is True:
        spec = importlib.util.spec_from_file_location("files", receiver_path)
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
        if 'class' in receiver_check[1]:
            try:
                target_class = getattr(module, receiver_check[1]['class'])
                instance = target_class()
                if hasattr(instance, receiver_check[1]['method']):
                    result = getattr(instance, receiver_check[1]['method'])(data)
                    return result
                else:
                    print(f"{type(receiver_check[1]).get('class')} class don't have {receiver_check[1]['method']} method")
            except AttributeError:
                print(f"class {type(receiver_check[1]).get('class')} can not find {receiver_path}")
        else:
            if hasattr(module, receiver_check[1]['method']):
                result = getattr(module, receiver_check[1]['method'])(data)
                return result
            else:
                print(f"{receiver_path} don't have {receiver_check[1]['method']} method")
    else:
        return
    
def receiver_server_callback(data, method) -> dict:
    if settings.PROJECT_NAME == 'bomiot':
        receiver_path = join(settings.BASE_DIR, 'core', 'server.py')
    else:
        receiver_path = join(settings.WORKING_SPACE, settings.PROJECT_NAME, 'server.py')
    receiver_check = check_method_in_file_by_ast(receiver_path, method)
    if receiver_check[0] is True:
        spec = importlib.util.spec_from_file_location("server", receiver_path)
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
        if 'class' in receiver_check[1]:
            try:
                target_class = getattr(module, receiver_check[1]['class'])
                instance = target_class()
                if hasattr(instance, receiver_check[1]['method']):
                    result = getattr(instance, receiver_check[1]['method'])(data)
                    return result
                else:
                    print(f"{type(receiver_check[1]).get('class')} class don't have {receiver_check[1]['method']} method")
            except AttributeError:
                print(f"class {type(receiver_check[1]).get('class')} can not find {receiver_path}")
        else:
            if hasattr(module, receiver_check[1]['method']):
                result = getattr(module, receiver_check[1]['method'])(data)
                return result
            else:
                print(f"{receiver_path} don't have {receiver_check[1]['method']} method")
    else:
        return
    
async def async_write_file(file_path, file_data):
    """
    Async write file
    """
    try:
        Path(file_path).parent.mkdir(parents=True, exist_ok=True)
        async with aiofiles.open(file_path, 'wb') as f:
            await f.write(file_data)
        return True
    except Exception as e:
        print(f"Error writing file {file_path}: {str(e)}")
        return False

def sync_write_file(file_path, file_data):
    """
    Sync use asyncio to async write file
    """
    try:
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        result = loop.run_until_complete(async_write_file(file_path, file_data))
        return result
    finally:
        loop.close()


from django.contrib.auth import get_user_model
from rest_framework.exceptions import APIException
from bomiot_message import login_message_return
from bomiot.server.core.jwt_auth import parse_payload
import typing

User = get_user_model()


class CoreAuthentication:
    """
    custom authenticate
    """

    def authenticate(self, request) -> typing.Tuple[bool, typing.Union[User, None]]:
        """
        auth user request
        :param request: HTTP request
        :return: (auth status, user)
        """
        # path that does not require authentication
        if request.path in ['/', '/django/api/docs/', '/django/api/debug/', '/django/api/', '/django/core/user/permission/']:
            return False, None

        # get request Token
        token = request.META.get('HTTP_TOKEN', '')
        if not token:
            self._raise_api_exception(request, 'Please Login First')

        # parse Token
        result = parse_payload(token)
        if result.get('status') is False:
            self._raise_api_exception(request, 'Please Login Again')
        user_id = result.get('data', {}).get('id')
        user_permissions = result.get('data', {}).get('permission', {})

        # authenticate user
        user = User.objects.filter(id=user_id, is_delete=False).first()
        if not user:
            self._raise_api_exception(request, "User not exists")

        # authenticate user status
        if not user.is_active:
            self._raise_api_exception(request, 'User is not active')

        # authenticate user permissions
        if sorted(user.permission.items()) != sorted(user_permissions.items()):
            self._raise_api_exception(request, 'Please Login Again')

        return True, user

    def authenticate_header(self, request) -> None:
        """
        authenticate header
        :param request: HTTP request
        """
        pass

    @staticmethod
    def _raise_api_exception(request, message: str) -> None:
        """
        Raise APIException exception
        :param request: HTTP request
        :param message: exception message
        """
        language = request.META.get('HTTP_LANGUAGE', '')
        raise APIException(login_message_return(language, message))


from locust import HttpUser, task, between

class PerformanceAPI(HttpUser):
    wait_time = between(1, 2)

    def on_start(self):
        self.login()

    def login(self):
        response = self.client.post(
            "/login/",
            json={"username": "admin", "password": "admin"}
        )
        result = response.json()
        self.token = result.get("token", "")
        self.headers = {"token": f"{self.token}"} if self.token else {}

    @task(1)
    def get_django_test(self):
        self.client.get("/test/", headers=self.headers, name="Get django test")

    @task(1)
    def get_fastapi_test(self):
        self.client.get("/fastapi/test/", headers=self.headers, name="Get fastapi test")
    
    @task(1)
    def get_flask_test(self):
        self.client.get("/flask/test/", headers=self.headers, name="Get flask test")


from fastapi import FastAPI,WebSocket
from django.contrib.auth import get_user_model

User = get_user_model()

fastapi_app = FastAPI(title="FastAPI APP")

@fastapi_app.get("/test/")
async def test():
    return {"msg": "This is FastAPI API"}

@fastapi_app.websocket("/ws")
async def websocket_endpoint(websocket: WebSocket):
    await websocket.accept()
    while True:
        try:
            data = await websocket.receive_text()
            await websocket.send_text(f"Message text was: {data}")
        except:
            break


from bomiot.server.core.jwt_auth import parse_payload
from bomiot_message import login_message_return
from rest_framework.exceptions import APIException
from django.utils.deprecation import MiddlewareMixin


class JwtAuthorizationMiddleware(MiddlewareMixin):
    """
    JWT Middleware
    """
    # allowed paths and prefixes
    ALLOWED_PATHS = [
        '/',
        '/login/',
        '/favicon.ico',
    ]
    ALLOWED_PREFIXES = [
        '/admin/',
        '/statics/',
        '/js/',
        '/css/',
        '/assets/',
    ]
    ALLOWED_METHODS = ['OPTIONS']

    def process_request(self, request):
        """
        auth request JWT Token
        :param request: HTTP request
        """
        if self._is_allowed_path(request) or request.method in self.ALLOWED_METHODS:
            return

        token = request.META.get('HTTP_TOKEN', '')
        if token:
            result = parse_payload(token).get('status')
            if result is True:
                return
            else:
                raise APIException(
                    login_message_return(request.META.get('HTTP_LANGUAGE', ''), 'Please Login Again')
                )
        else:
            raise APIException(
                login_message_return(request.META.get('HTTP_LANGUAGE', ''), 'Please Login First')
            )

    def _is_allowed_path(self, request):
        """
        check if the request path is allowed
        :param request: HTTP request
        :return: bool
        """
        if request.path_info in self.ALLOWED_PATHS:
            return True
        for prefix in self.ALLOWED_PREFIXES:
            if request.path_info.startswith(prefix):
                return True
        return False
    


from bomiot_message import msg_message_return, detail_message_return, login_message_return
from bomiot.server.core.models import Example
from bomiot.server.core.utils import queryset_to_dict
from django.core.cache import cache


# class ExampleClass(object):

#     def example_get(self, data):
#         print(data.get('query_params').get('params'))
#         example_list = Example.objects.filter()
#         qs_list = queryset_to_dict(example_list)
#         return [
#             ('results', data.get('data')),
#         ]

#     def example_create(self, data):
#         print(data.get('data'))
#         language = data.get('request').META.get('HTTP_LANGUAGE', 'en-US')
#         return msg_message_return(language, "Success Create")
    
#     def example_update(self, data):
#         print(data.get('data'))
#         print(data.get('updated_fields'))
#         language = data.get('request').META.get('HTTP_LANGUAGE', 'en-US')
#         return msg_message_return(language, "Success Update")
    
#     def example_delete(self, data):
#         print(data.get('data'))
#         language = data.get('request').META.get('HTTP_LANGUAGE', 'en-US')
#         return msg_message_return(language, "Success Delete")


import ast
import orjson
import time
from os.path import join, exists
from tomlkit import parse
import random

from collections import OrderedDict
from rest_framework.pagination import PageNumberPagination
from rest_framework.response import Response
from rest_framework.utils.urls import replace_query_param
from django.conf import settings
from django.core.cache import cache
from bomiot.server.core.utils import flatten_json, all_fields_empty
from bomiot.server.core.signal import bomiot_data_signals
from bomiot.server.core.models import Permission
from bomiot_message import permission_message_return

class CorePageNumberPagination(PageNumberPagination):
    page_query_param = 'page'
    page_size = 30
    page_size_query_param = "max_page"
    max_page_size = 1000

    def get_previous_link(self):
        """
        get previous link
        :return: previous link whole URL or None
        """
        if not self.page.has_previous():
            return None
        return self._build_absolute_url(self.page.previous_page_number())

    def get_next_link(self):
        """
        get next link
        :return: next link whole URL or None
        """
        if not self.page.has_next():
            return None
        return self._build_absolute_url(self.page.next_page_number())

    def _build_absolute_url(self, page_number):
        """
        resoleve absolute URL
        :param page_number: page number
        :return: full URL
        """
        url = self.request.build_absolute_uri()
        ssl_scheme = self.request.scheme
        url_parts = str(url).split(':', 1)

        if url_parts[0] == ssl_scheme:
            return replace_query_param(url, self.page_query_param, page_number)
        else:
            # Fix URL SSL scheme
            corrected_url = f"{ssl_scheme}:{url_parts[1]}"
            return replace_query_param(corrected_url, self.page_query_param, page_number)
        
    def query_data_add(self) -> list:
        return []

    def get_paginated_response(self, data):
        response_data = [
            ('count', self.page.paginator.count),
            ('next', self.get_next_link()),
            ('previous', self.get_previous_link()),
            ('results', data)
        ]
        response_data += self.query_data_add()
        return Response(OrderedDict(response_data))


class PermissionPageNumberPagination(PageNumberPagination):
    page_query_param = 'page'
    page_size = 1000
    page_size_query_param = "max_page"
    max_page_size = 10000
    time_ns = int(time.time_ns()) + random.randint(0, 10000)

    def get_previous_link(self):
        """
        get previous link
        :return: previous link whole URL or None
        """
        if not self.page.has_previous():
            return None
        return self._build_absolute_url(self.page.previous_page_number())

    def get_next_link(self):
        """
        get next link
        :return: next link whole URL or None
        """
        if not self.page.has_next():
            return None
        return self._build_absolute_url(self.page.next_page_number())

    def _build_absolute_url(self, page_number):
        """
        resoleve absolute URL
        :param page_number: page number
        :return: full URL
        """
        url = self.request.build_absolute_uri()
        ssl_scheme = self.request.scheme
        url_parts = str(url).split(':', 1)

        if url_parts[0] == ssl_scheme:
            return replace_query_param(url, self.page_query_param, page_number)
        else:
            # Fix URL SSL scheme
            corrected_url = f"{ssl_scheme}:{url_parts[1]}"
            return replace_query_param(corrected_url, self.page_query_param, page_number)
        
    def query_data_add(self) -> list:
        return []
    
    def get_permission_return_data(self, data) -> dict:
        permission_data = cache.get("permission_message", version=self.time_ns)
        if data.get('name') in permission_data:
            return {
                "label": permission_data.get(data.get('name')),
                "value": data.get('name')
            }
        else:
            return {
                "label": data.get('name'),
                "value": data.get('name')
            }

    def get_paginated_response(self, data):
        language = self.request.META.get('HTTP_LANGUAGE', 'en-US')
        message_path = join(settings.LANGUAGE_DIR, language + '.toml')
        message_dict = {}
        if exists(message_path):
            with open(message_path, 'r', encoding='utf-8') as message:
                message_data = parse(message.read())
        message_dict = message_data['permission']
        cache.set('permission_message', orjson.loads(orjson.dumps(message_dict).decode("utf-8")), version=self.time_ns)
        data_list = list(map(lambda x: self.get_permission_return_data(x), data))
        cache.delete("permission_message", version=self.time_ns)
        response_data = [
            ('count', self.page.paginator.count),
            ('next', self.get_next_link()),
            ('previous', self.get_previous_link()),
            ('results', data_list)
        ]
        response_data += self.query_data_add()
        return Response(OrderedDict(response_data))


class TeamPageNumberPagination(PageNumberPagination):
    page_query_param = 'page'
    page_size = 30
    page_size_query_param = "max_page"
    max_page_size = 1000
    time_ns = int(time.time_ns()) + random.randint(0, 10000)

    def get_previous_link(self):
        """
        get previous link
        :return: previous link whole URL or None
        """
        if not self.page.has_previous():
            return None
        return self._build_absolute_url(self.page.previous_page_number())

    def get_next_link(self):
        """
        get next link
        :return: next link whole URL or None
        """
        if not self.page.has_next():
            return None
        return self._build_absolute_url(self.page.next_page_number())

    def _build_absolute_url(self, page_number):
        """
        resoleve absolute URL
        :param page_number: page number
        :return: full URL
        """
        url = self.request.build_absolute_uri()
        ssl_scheme = self.request.scheme
        url_parts = str(url).split(':', 1)

        if url_parts[0] == ssl_scheme:
            return replace_query_param(url, self.page_query_param, page_number)
        else:
            # Fix URL SSL scheme
            corrected_url = f"{ssl_scheme}:{url_parts[1]}"
            return replace_query_param(corrected_url, self.page_query_param, page_number)
        
    def query_data_add(self) -> list:
        return []
    
    def get_permission_return_data(self, data) -> dict:
        permission_message = cache.get("permission_message", version=self.time_ns)
        if data.name in permission_message:
            return {
                "label": permission_message.get(data.name),
                "value": data.name
            }
        else:
            return {
                "label": data.name,
                "value": data.name
            }

    def get_paginated_response(self, data):
        language = self.request.META.get('HTTP_LANGUAGE', 'en-US')
        message_path = join(settings.LANGUAGE_DIR, language + '.toml')
        message_dict = {}
        if exists(message_path):
            with open(message_path, 'r', encoding='utf-8') as message:
                message_data = parse(message.read())
        message_dict = message_data['permission']
        cache.set('permission_message', orjson.loads(orjson.dumps(message_dict).decode("utf-8")), version=self.time_ns)
        permission_list = Permission.objects.filter()
        permission_data_list = list(map(lambda x: self.get_permission_return_data(x), permission_list))
        cache.delete("permission_message", version=self.time_ns)
        response_data = [
            ('count', self.page.paginator.count),
            ('next', self.get_next_link()),
            ('previous', self.get_previous_link()),
            ('permission_list', permission_data_list),
            ('results', data)
        ]
        response_data += self.query_data_add()
        return Response(OrderedDict(response_data))


class DataCorePageNumberPagination(PageNumberPagination):
    page_query_param = 'page'
    page_size = 30
    page_size_query_param = "max_page"
    max_page_size = 1000

    def get_previous_link(self):
        """
        get previous link
        :return: previous link whole URL or None
        """
        if not self.page.has_previous():
            return None
        return self._build_absolute_url(self.page.previous_page_number())

    def get_next_link(self):
        """
        get next link
        :return: next link whole URL or None
        """
        if not self.page.has_next():
            return None
        return self._build_absolute_url(self.page.next_page_number())

    def _build_absolute_url(self, page_number):
        """
        resoleve absolute URL
        :param page_number: page number
        :return: full URL
        """
        url = self.request.build_absolute_uri()
        ssl_scheme = self.request.scheme
        url_parts = str(url).split(':', 1)

        if url_parts[0] == ssl_scheme:
            return replace_query_param(url, self.page_query_param, page_number)
        else:
            # Fix URL SSL scheme
            corrected_url = f"{ssl_scheme}:{url_parts[1]}"
            return replace_query_param(corrected_url, self.page_query_param, page_number)
        
    def get_paginated_response(self, data):
        response_data = [
            ('count', self.page.paginator.count),
            ('next', self.get_next_link()),
            ('previous', self.get_previous_link()),
        ]
        data_list = list(map(lambda x: flatten_json(x) if isinstance(x, dict) else x, data))
        origin = self.request.query_params.dict()
        params_check = self.request.query_params.get('params', {})
        if params_check:
            params_str = params_check.replace('true', 'True').replace('false', 'False')
            params_str = params_check.replace("'", '"')
            params_dict = ast.literal_eval(params_str)
            if all_fields_empty(params_dict):
                origin['params'] = {}
            origin['params'] = params_dict
        else:
            origin['params'] = {}
        responses = bomiot_data_signals.send_robust(sender=self.__class__,
                                                    request=self.request,
                                                    mode='get',
                                                    query_params=origin,
                                                    data=data_list)
        for receiver, response in responses:
            if isinstance(response, Exception):
                raise response
            if isinstance(response, list):
                callback_data = True
                for i in response:
                    if i[0] == 'results':
                        callback_data = False
                        break
                    else:
                        continue
                if callback_data is False:
                    response_data += [('results', data_list)]
                response_data += response
        return Response(OrderedDict(response_data))


class APIPageNumberPagination(PageNumberPagination):
    page_query_param = 'page'
    page_size = 30
    page_size_query_param = "max_page"
    max_page_size = 1000
    time_ns = int(time.time_ns()) + random.randint(0, 10000)

    def get_previous_link(self):
        """
        get previous link
        :return: previous link whole URL or None
        """
        if not self.page.has_previous():
            return None
        return self._build_absolute_url(self.page.previous_page_number())

    def get_next_link(self):
        """
        get next link
        :return: next link whole URL or None
        """
        if not self.page.has_next():
            return None
        return self._build_absolute_url(self.page.next_page_number())

    def _build_absolute_url(self, page_number):
        """
        resoleve absolute URL
        :param page_number: page number
        :return: full URL
        """
        url = self.request.build_absolute_uri()
        ssl_scheme = self.request.scheme
        url_parts = str(url).split(':', 1)

        if url_parts[0] == ssl_scheme:
            return replace_query_param(url, self.page_query_param, page_number)
        else:
            # Fix URL SSL scheme
            corrected_url = f"{ssl_scheme}:{url_parts[1]}"
            return replace_query_param(corrected_url, self.page_query_param, page_number)
        
    def query_data_add(self) -> list:
        return []

    def get_api_return_data(self, data) -> dict:
        permission_data = cache.get("permission_message", version=self.time_ns)
        if data.get('name') in permission_data:
            data['name'] = permission_data.get(data.get('name'))
        return data
    
    def get_paginated_response(self, data):
        language = self.request.META.get('HTTP_LANGUAGE', 'en-US')
        message_path = join(settings.LANGUAGE_DIR, language + '.toml')
        message_dict = {}
        if exists(message_path):
            with open(message_path, 'r', encoding='utf-8') as message:
                message_data = parse(message.read())
        message_dict = message_data['permission']
        cache.set('permission_message', orjson.loads(orjson.dumps(message_dict).decode("utf-8")), version=self.time_ns)
        data_list = list(map(lambda x: self.get_api_return_data(x), data))
        cache.delete("permission_message", version=self.time_ns)
        response_data = [
            ('count', self.page.paginator.count),
            ('next', self.get_next_link()),
            ('previous', self.get_previous_link()),
            ('results', data_list)
        ]
        response_data += self.query_data_add()
        return Response(OrderedDict(response_data))


import subprocess
import socket

def kill_process_on_port(port: int):
    def is_port_in_use(port):
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            return s.connect_ex(('localhost', port)) == 0
    
    def kill_process(port):
        try:
            result = subprocess.run(
                f'netstat -ano | findstr :{port}',
                shell=True, capture_output=True, text=True
            )
            if result.stdout:
                lines = result.stdout.strip().split('\n')
                for line in lines:
                    if f':{port}' in line:
                        parts = line.split()
                        if len(parts) >= 5:
                            pid = parts[-1]
                            if pid != 0:
                                subprocess.run(f'taskkill /PID {pid} /F', shell=True)
        except Exception as e:
            print(f"Error: {e}")
    
    if is_port_in_use(port):
        kill_process(port)




# class ServerClass:
    # def pid_get(self, data):
    #     print(data)
    
    # def network_get(self, data):
    #     print(data)

    # def disk_get(self, data):
    #     print(data)

    # def cpu_get(self, data):
    #     print(data)

    # def memory_get(self, data):
    #     print(data)


import sys
import os
import uvicorn
import time
from bomiot import version
import argparse
import psutil
import multiprocessing
import threading

optional_title = 'Optional arguments'


class CapitalisedHelpFormatter(argparse.HelpFormatter):
    def __init__(self, prog):
        super(CapitalisedHelpFormatter, self).__init__(prog,
                                                       indent_increment=2,
                                                       max_help_position=30,
                                                       width=200)
        self._action_max_length = 20

    def add_usage(self, usage, actions, groups, prefix=None):
        if prefix is None:
            prefix = 'Usage: '
        return super(CapitalisedHelpFormatter, self).add_usage(
            usage, actions, groups, prefix)

    class _Section(object):

        def __init__(self, formatter, parent, heading=None):
            self.formatter = formatter
            self.parent = parent
            self.heading = heading
            self.items = []

        def format_help(self):
            # format the indented section
            if self.parent is not None:
                self.formatter._indent()
            join = self.formatter._join_parts
            item_help = join([func(*args) for func, args in self.items])
            if self.parent is not None:
                self.formatter._dedent()

            # return nothing if the section was empty
            if not item_help:
                return ''

            # add the heading if the section was non-empty
            if self.heading is not argparse.SUPPRESS and self.heading is not None:
                current_indent = self.formatter._current_indent
                if self.heading == optional_title:
                    heading = '%*s%s:\n' % (current_indent, '', self.heading)
                else:
                    heading = '%*s%s:' % (current_indent, '', self.heading)
            else:
                heading = ''

            return join(['\n', heading, item_help])


parser = argparse.ArgumentParser(description='Bomiot %s - Distributed Crawler Management Framework' % version(),
                                 formatter_class=CapitalisedHelpFormatter, add_help=False)
parser._optionals.title = optional_title

parser.add_argument('-v', '--version', action='version',
                    version=version(), help='Get version of Bomiot')
parser.add_argument('-h', '--help', action='help',
                    help='Show this help message and exit')

subparsers = parser.add_subparsers(
    dest='command', title='Available commands', metavar='')

# new
parser_new = subparsers.add_parser(
    'new', help='Create APP for project')
parser_new.add_argument('folder', default='',
                         nargs='?', type=str, help='Create APP')

# project
parser_project = subparsers.add_parser(
    'project', help='Project workspace, default to bomiot')
parser_project.add_argument('folder', default='',
                         nargs='?', type=str, help='project workspace folder')

# plugins
parser_plugins = subparsers.add_parser(
    'plugins', help='Plugins workspace, default to bomiot')
parser_plugins.add_argument('folder', default='',
                         nargs='?', type=str, help='plugins workspace folder')

# deploy
parser_deploy = subparsers.add_parser(
    'deploy', help='Deploy project')
parser_deploy.add_argument('folder', default='',
                         nargs='?', type=str, help='deploy project')

# init
parser_init = subparsers.add_parser(
    'init', help='Init bomiot')
parser_init.add_argument('folder', default='',
                         nargs='?', type=str, help='Init Workingspace')

# init admin
parser_initadmin = subparsers.add_parser(
    'initadmin', help='Create default super user admin')

# init password
parser_initpwd = subparsers.add_parser(
    'initpwd', help='Init admin password')

# migrate
parser_migrate = subparsers.add_parser('migrate', help='Migrate database')

# makemigrations
parser_makemigrations = subparsers.add_parser(
    'makemigrations', help='Generate migrations for database')
parser_makemigrations.add_argument('appname', nargs='?', type=str, help='App name (e.g., xx or xxx.xx)')
parser_makemigrations.add_argument('--empty', action='store_true', help='Create an empty migration')
parser_makemigrations.add_argument('--dry-run', action='store_true', help='Show what would be done without making changes')

# loaddata
parser_loaddata = subparsers.add_parser(
    'loaddata', help='Load data from configs')
parser_loaddata.add_argument('source', type=str, help='Configs path')

# dumpdata
parser_dumpdata = subparsers.add_parser(
    'dumpdata', help='Dump data to configs')
parser_dumpdata.add_argument(
    'appname', default='core', nargs='?', type=str, help='Appname')

# marketplace
parser_marketplace = subparsers.add_parser(
    'market', help='Copy project from marketplace')
parser_marketplace.add_argument('folder', default='',
                         nargs='?', type=str, help='project marketplace folder')

# init auth keys
parser_keys = subparsers.add_parser(
    'keys', help='Init auth keys')

# run
parser_run = subparsers.add_parser(
    'run', help='Run server')
parser_run.add_argument("--host", "-b", type=str, default="127.0.0.1", help="Default Domin: 127.0.0.1")
parser_run.add_argument("--port", "-p", type=int, default=8000, help="Default Pore: 8000")
parser_run.add_argument("--workers", "-w", type=int, default=1, help="CPU Core")
parser_run.add_argument("--log-level", type=str, default="info", choices=["critical", "error", "warning", "info", "debug", "trace"], help="Log Level")
parser_run.add_argument("--uds", type=str, default=None, help="UNIX domain socket")
parser_run.add_argument("--ssl-keyfile", type=str, default=None, help="SSL Key")
parser_run.add_argument("--ssl-certfile", type=str, default=None, help="SSL PEM")
parser_run.add_argument("--proxy-headers", action="store_true", help="X-Forwarded-*Header")
parser_run.add_argument("--http", type=str, default="httptools", choices=["auto", "h11", "httptools"], help="HTTP")
parser_run.add_argument("--loop", "-l", type=str, default="auto", choices=["auto", "asyncio", "uvloop"], help="Asyncio Loop")
parser_run.add_argument("--limit-concurrency", type=str, default=1000, help="Limit concurrency")
parser_run.add_argument("--backlog", type=int, default=128, help="Backlog")
parser_run.add_argument("--timeout-keep-alive", type=str, default=5, help="Backlog")
parser_run.add_argument("--timeout-graceful-shutdown", type=str, default=30, help="Time out shut down")
parser_run.add_argument("--server-header", type=bool, default=False, help="Server Header")
parser_run.add_argument("--app", type=str, default="bomiot_asgi:application", help="ASGI Application")

# show help info when no args
if len(sys.argv[1:]) == 0:
    parser.print_help()
    parser.exit()


def cmd():
    """
    run from cmd
    :return:
    """
    args = parser.parse_args()
    command = args.command
    # project workspace for bomiot
    if command == 'project':
        from bomiot.cmd.project import project
        project(args.folder)
    # plugins workspace for bomiot
    elif command == 'plugins':
        from bomiot.cmd.plugins import plugins
        plugins(args.folder)
    # deploy project
    elif command == 'deploy':
        from bomiot.cmd.deploy import deploy
        deploy(args.folder)
    # init bomiot
    elif command == 'init':
        from bomiot.cmd.init import create_file
        create_file(args.folder)
    # init admin
    elif command == 'initadmin':
        from bomiot.cmd.initadmin import init_admin
        init_admin()
    # init admin
    elif command == 'initpwd':
        from bomiot.cmd.initpwd import init_password
        init_password()
    # create app
    elif command == 'new':
        from bomiot.cmd.createapp import new_app
        new_app(args.folder)
    # marketplace
    elif command == 'market':
        from bomiot.cmd.market import copy_project
        copy_project(args.folder)
    # init auth keys
    elif command == 'keys':
        from bomiot.cmd.create_key import auth_key_refresh
        auth_key_refresh()
    # makemigrations
    elif command == 'makemigrations':
        os.environ.setdefault("DJANGO_SETTINGS_MODULE", "bomiot.server.server.settings")
        os.environ.setdefault('RUN_MAIN', 'true')
        import django
        django.setup()
        from django.core.management import call_command
        from django.apps import apps
        appname = args.appname
        empty = args.empty
        dry_run = args.dry_run
        cmd_args = ['makemigrations']
        if appname:
            if '.' in appname:
                found_app_label = None
                for app_config in apps.get_app_configs():
                    if app_config.name == appname:
                        found_app_label = app_config.label
                        break
                if found_app_label:
                    cmd_args.append(found_app_label)
                else:
                    return
            else:
                found_app = None
                for app_config in apps.get_app_configs():
                    if app_config.label == appname or app_config.name.endswith('.' + appname):
                        found_app = app_config.label
                        break
                if found_app:
                    cmd_args.append(found_app)
                else:
                    return
        else:
            apps_with_models = []
            for app_config in apps.get_app_configs():
                try:
                    if app_config.models_module:
                        models = apps.get_app_config(app_config.label).get_models()
                        if models:
                            apps_with_models.append(app_config.label)
                except Exception:
                    continue
            if apps_with_models:
                cmd_args.extend(apps_with_models)
            else:
                return
        if empty:
            cmd_args.append('--empty')
        if dry_run:
            cmd_args.append('--dry-run')
        try:
            call_command(*cmd_args)
        except Exception as e:
            print(f"Error creating migrations: {e}")
    # migrate
    elif command == 'migrate':
        from bomiot.cmd.migrate import migrate
        migrate()
    # run server
    elif command == 'run':
        import platform
        import importlib.resources
        from os.path import join
        from pathlib import Path
        from configparser import ConfigParser
        workspace_path = importlib.resources.files('bomiot').joinpath('server', 'workspace.ini')
        WORKING_SPACE_CONFIG = ConfigParser()
        WORKING_SPACE_CONFIG.read(workspace_path, encoding='utf-8')
        WORKING_SPACE = WORKING_SPACE_CONFIG.get('space', 'name', fallback='Create your working space first')
        if platform.system() == 'Windows':
            from bomiot.cmd.killport import kill_process_on_port
            kill_process_on_port(args.port)
        lockfile = Path(join(WORKING_SPACE, 'bomiot_ready.lock'))
        if lockfile.exists():
            lockfile.unlink()
        os.environ.setdefault("DJANGO_SETTINGS_MODULE", "bomiot.server.server.settings")
        os.environ.setdefault('RUN_MAIN', 'true')
        os.environ.setdefault('WORKERS', str(args.workers))
        uvicorn.run(
            args.app,
            host=args.host,
            port=args.port,
            workers=args.workers,
            log_level=args.log_level,
            uds=args.uds,
            ssl_keyfile=args.ssl_keyfile,
            ssl_certfile=args.ssl_certfile,
            proxy_headers=args.proxy_headers,
            http=args.http,
            server_header=False,
            limit_concurrency=args.limit_concurrency,
            backlog=args.backlog,
            timeout_keep_alive=args.timeout_keep_alive,
            timeout_graceful_shutdown=args.timeout_graceful_shutdown,
            loop=args.loop
        )


# for console debugger
if __name__ == '__main__':
    cmd()



import django
import os


def init_password():
    """
    init superuser password
    :return:
    """
    os.environ.setdefault("DJANGO_SETTINGS_MODULE", "bomiot.server.server.settings")
    django.setup()
    from django.contrib.auth import get_user_model

    User = get_user_model()

    user_data = User.objects.filter(username='admin', is_superuser=True)
    if user_data.exists:
        user_data.first().set_password('admin')
        print('Reset admin account: admin, initial password: admin, just use it temporarily '
              'and change the password for safety')
    else:
        print('Please init one admin first')
        


import re
from bomiot_token import encrypt_info

from os.path import join, exists 
from os import getcwd

def auth_key_get():
    path = join(getcwd(), 'auth_key.py')
    if not exists(join(path)):
        while True:
            key_code = encrypt_info()
            if '/' in key_code:
                continue
            else:
                break
        with open(path, "w", encoding="utf-8") as f:
            f.write(f'KEY = "{key_code}"\n')
        f.close()
        print(f'keys: {key_code}')

def auth_key_refresh():
    path = join(getcwd(), 'auth_key.py')
    if not exists(join(path)):
        while True:
            key_code = encrypt_info()
            if '/' in key_code:
                continue
            else:
                break
        with open(path, "w", encoding="utf-8") as f:
            f.write(f'KEY = "{key_code}"\n')
        f.close()
        print(f'keys: {key_code}')
    else:
        while True:
            new_key = encrypt_info()
            if '/' in new_key:
                continue
            else:
                break
        with open(path, 'r', encoding='utf-8') as f:
            lines = f.readlines()
        pattern = re.compile(rf"^(\s*KEY\s*=\s*)(.*)$")
        new_lines = []
        for line in lines:
            print(pattern.match(line))
            if pattern.match(line):
                prefix = pattern.match(line).group(1)
                if isinstance(new_key, str) and not new_key.isdigit():
                    value_str = f'"{new_key}"'
                else:
                    value_str = str(new_key)
                new_line = f"{prefix}{value_str}\n"
                new_lines.append(new_line)
        if len(new_lines) == 0:
            if isinstance(new_key, str) and not new_key.isdigit():
                value_str = f'"{new_key}"'
            else:
                value_str = str(new_key)
            new_lines.append(f'KEY = {value_str}\n')
        with open(path, 'w', encoding='utf-8') as f:
            f.writelines(new_lines)
        f.close()
        print(f'new keys: {new_key}')


def auth_key_force():
    path = join(getcwd(), 'auth_key.py')
    if not exists(join(path)):
        while True:
            key_code = encrypt_info()
            if '/' in key_code:
                continue
            else:
                break
        with open(path, "w", encoding="utf-8") as f:
            f.write(f'KEY = "{key_code}"\n')
        f.close()
        return key_code
    else:
        while True:
            new_key = encrypt_info()
            if '/' in new_key:
                continue
            else:
                break
        with open(path, 'r', encoding='utf-8') as f:
            lines = f.readlines()
        pattern = re.compile(rf"^(\s*KEY\s*=\s*)(.*)$")
        new_lines = []
        for line in lines:
            if pattern.match(line):
                prefix = pattern.match(line).group(1)
                if isinstance(new_key, str) and not new_key.isdigit():
                    value_str = f'"{new_key}"'
                else:
                    value_str = str(new_key)
                new_line = f"{prefix}{value_str}\n"
                new_lines.append(new_line)
        if len(new_lines) == 0:
            if isinstance(new_key, str) and not new_key.isdigit():
                value_str = f'"{new_key}"'
            else:
                value_str = str(new_key)
            new_lines.append(f'KEY = {value_str}\n')
        with open(path, 'w', encoding='utf-8') as f:
            f.writelines(new_lines)
        f.close()
        return new_key



from django.conf import settings
from django.contrib.auth.models import AbstractUser
from django.db import models


class CoreModel(models.Model):
    is_delete = models.BooleanField(default=False, verbose_name='Delete Label')
    created_time = models.DateTimeField(auto_now_add=True, verbose_name="Created Time")
    updated_time = models.DateTimeField(auto_now=True, blank=True, null=True, verbose_name="Updated Time")

    class Meta:
        ordering = ['-id']
        abstract = True


class User(AbstractUser, CoreModel):
    type = models.IntegerField(default=1, verbose_name="User Type")
    phone = models.CharField(default='', max_length=255, blank=True, verbose_name="Phone")
    permission = models.JSONField(default=dict, null=True, verbose_name="Permission")
    request_limit = models.IntegerField(default=0, verbose_name="Request Limit")
    team = models.IntegerField(default=0, blank=True, verbose_name="Team")
    department = models.IntegerField(default=0, blank=True, verbose_name="Department")

    class Meta(AbstractUser.Meta):
        db_table = settings.BASE_DB_TABLE + '_user'
        verbose_name = settings.BASE_DB_TABLE + ' User'
        verbose_name_plural = verbose_name
        ordering = ['-id']


class Permission(CoreModel):
    api = models.CharField(max_length=255, verbose_name="Permission API")
    name = models.CharField(max_length=255, verbose_name="Permission Name")

    class Meta:
        db_table = settings.BASE_DB_TABLE + '_permission'
        verbose_name = settings.BASE_DB_TABLE + ' Permission'
        verbose_name_plural = verbose_name
        ordering = ['-id']


class API(CoreModel):
    id = models.AutoField(primary_key=True)
    method = models.CharField(max_length=18, verbose_name="Method", default='GET')
    api = models.CharField(max_length=255, verbose_name="API API")
    func_name = models.CharField(max_length=255, verbose_name="API Name")
    name = models.CharField(max_length=255, verbose_name="API Description", default='')

    def __str__(self):
        return self.api

    class Meta:
        db_table = settings.BASE_DB_TABLE + '_api'
        verbose_name = settings.BASE_DB_TABLE + ' API'
        verbose_name_plural = verbose_name
        ordering = ['id']


class ThrottleModel(CoreModel):
    ip = models.CharField(max_length=255, verbose_name="IP")
    method = models.CharField(max_length=18, verbose_name="Method")

    class Meta:
        db_table = settings.BASE_DB_TABLE + '_throttle'
        verbose_name = settings.BASE_DB_TABLE + ' Throttle'
        verbose_name_plural = verbose_name
        ordering = ['-id']


class JobList(CoreModel):
    job_id = models.CharField(max_length=255, verbose_name="Job ID")
    module_name = models.CharField(max_length=255, verbose_name="Module Name")
    func_name = models.CharField(max_length=255, verbose_name="Function Name")
    trigger = models.CharField(max_length=255, verbose_name="Trigger")
    description = models.TextField(default='', null=True, blank=True, verbose_name="Description")
    configuration = models.TextField(null=True, verbose_name="Configuration")
    type = models.BooleanField(default=True, verbose_name="Type")

    class Meta:
        db_table = settings.BASE_DB_TABLE + '_job'
        verbose_name = settings.BASE_DB_TABLE + ' Job'
        verbose_name_plural = verbose_name
        ordering = ['-id']


class Files(CoreModel):
    name = models.CharField(default='', max_length=255, blank=True, verbose_name="File Name")
    type = models.CharField(default='', max_length=255, blank=True, verbose_name="Type")
    size = models.BigIntegerField(default=0, blank=True, verbose_name="Size")
    owner = models.CharField(default='', max_length=255, blank=True, verbose_name="Owner")
    shared_to = models.CharField(default='', max_length=255, blank=True, verbose_name="Shared To")

    class Meta:
        db_table = settings.BASE_DB_TABLE + '_files'
        verbose_name = settings.BASE_DB_TABLE + ' Files'
        verbose_name_plural = verbose_name
        ordering = ['-updated_time']


class Message(CoreModel):
    sender = models.CharField(default='', max_length=255, blank=True, verbose_name="Sender")
    receiver = models.CharField(default='', max_length=255, blank=True, verbose_name="Receiver")
    detail = models.CharField(default='', max_length=255, verbose_name="Detail")
    can_send = models.BooleanField(default=False, verbose_name="Can Send")

    class Meta:
        db_table = settings.BASE_DB_TABLE + '_message'
        verbose_name = settings.BASE_DB_TABLE + ' Message'
        verbose_name_plural = verbose_name
        ordering = ['-id']


class Pids(CoreModel):
    pid = models.IntegerField(default=0, verbose_name="PID")
    name = models.CharField(default='', max_length=255, blank=True, verbose_name="PID Name")
    memory = models.BigIntegerField(default=0, verbose_name="Memory")
    create_time = models.DateTimeField(auto_now_add=False, auto_now=False, blank=True, verbose_name="Create Time")
    memory_usage = models.FloatField(default=0, verbose_name="Memory Usage")
    cpu_usage = models.FloatField(default=0, verbose_name="CPU Usage")

    class Meta:
        db_table = settings.BASE_DB_TABLE + '_pids'
        verbose_name = settings.BASE_DB_TABLE + ' Pids'
        verbose_name_plural = verbose_name
        ordering = ['-cpu_usage']


class CPU(CoreModel):
    cpu_usage = models.FloatField(default=0, verbose_name="CPU Usage")
    physical_cores = models.IntegerField(default=0, verbose_name="Physical Cores")
    logical_cores = models.IntegerField(default=0, verbose_name="Logical Cores")
    cpu_frequency = models.CharField(default='', max_length=255, verbose_name="CPU Frequency")
    min_cpu_frequency = models.CharField(default='', max_length=255, verbose_name="Min CPU Frequency")
    max_cpu_frequency = models.CharField(default='', max_length=255, verbose_name="Max CPU Frequency")

    class Meta:
        db_table = settings.BASE_DB_TABLE + '_cpu'
        verbose_name = settings.BASE_DB_TABLE + ' CPU'
        verbose_name_plural = verbose_name
        ordering = ['-id']


class Memory(CoreModel):
    total = models.BigIntegerField(default=0, verbose_name="Total Memory")
    used = models.BigIntegerField(default=0, verbose_name="Used Memory")
    free = models.BigIntegerField(default=0, verbose_name="Free Memory")
    percent = models.FloatField(default=0, verbose_name="Percent")
    swap_total = models.BigIntegerField(default=0, verbose_name="Swap Total Memory")
    swap_used = models.BigIntegerField(default=0, verbose_name="Swap Used Memory")
    swap_free = models.BigIntegerField(default=0, verbose_name="Swap Free Memory")
    swap_percent = models.FloatField(default=0, verbose_name="Swap Percent")

    class Meta:
        db_table = settings.BASE_DB_TABLE + '_memory'
        verbose_name = settings.BASE_DB_TABLE + ' Memory'
        verbose_name_plural = verbose_name
        ordering = ['-id']


class Disk(CoreModel):
    device = models.CharField(default='', max_length=255, verbose_name="Device")
    mountpoint = models.CharField(default='', max_length=255, verbose_name="Mountpoint")
    total = models.BigIntegerField(default=0, verbose_name="Total")
    used = models.BigIntegerField(default=0, verbose_name="Used")
    free = models.BigIntegerField(default=0, verbose_name="Free")
    percent = models.FloatField(default=0, verbose_name="Percent")

    class Meta:
        db_table = settings.BASE_DB_TABLE + '_disk'
        verbose_name = settings.BASE_DB_TABLE + ' Disk'
        verbose_name_plural = verbose_name
        ordering = ['-id']


class Network(CoreModel):
    bytes_sent = models.BigIntegerField(default=0, verbose_name="Bytes Sent")
    bytes_recv = models.BigIntegerField(default=0, verbose_name="Bytes Received")

    class Meta:
        db_table = settings.BASE_DB_TABLE + '_network'
        verbose_name = settings.BASE_DB_TABLE + ' Network'
        verbose_name_plural = verbose_name
        ordering = ['-id']


class Team(CoreModel):
    name = models.CharField(default='', max_length=255, verbose_name="Team Name")
    permission = models.JSONField(default=dict, null=True, verbose_name="Permission")

    class Meta:
        db_table = settings.BASE_DB_TABLE + '_team'
        verbose_name = settings.BASE_DB_TABLE + ' Team'
        verbose_name_plural = verbose_name
        ordering = ['-id']


class Example(CoreModel):
    data = models.JSONField()

    class Meta:
        db_table = settings.BASE_DB_TABLE + '_example'
        verbose_name = settings.BASE_DB_TABLE + ' Example'
        verbose_name_plural = verbose_name
        ordering = ['-id']


class Department(CoreModel):
    name = models.CharField(default='', max_length=255, verbose_name="Department Name")

    class Meta:
        db_table = settings.BASE_DB_TABLE + '_department'
        verbose_name = settings.BASE_DB_TABLE + ' Department'
        verbose_name_plural = verbose_name
        ordering = ['-id']


class Goods(CoreModel):
    data = models.JSONField()

    class Meta:
        db_table = settings.BASE_DB_TABLE + '_goods'
        verbose_name = settings.BASE_DB_TABLE + ' Goods'
        verbose_name_plural = verbose_name
        ordering = ['-id']


class Bin(CoreModel):
    data = models.JSONField()

    class Meta:
        db_table = settings.BASE_DB_TABLE + '_bin'
        verbose_name = settings.BASE_DB_TABLE + ' Bin'
        verbose_name_plural = verbose_name
        ordering = ['-id']


class Stock(CoreModel):
    data = models.JSONField()

    class Meta:
        db_table = settings.BASE_DB_TABLE + '_stock'
        verbose_name = settings.BASE_DB_TABLE + ' Stock'
        verbose_name_plural = verbose_name
        ordering = ['-id']


class Capital(CoreModel):
    data = models.JSONField()

    class Meta:
        db_table = settings.BASE_DB_TABLE + '_capital'
        verbose_name = settings.BASE_DB_TABLE + ' Capital'
        verbose_name_plural = verbose_name
        ordering = ['-id']


class Supplier(CoreModel):
    data = models.JSONField()

    class Meta:
        db_table = settings.BASE_DB_TABLE + '_supplier'
        verbose_name = settings.BASE_DB_TABLE + ' Supplier'
        verbose_name_plural = verbose_name
        ordering = ['-id']


class Customer(CoreModel):
    data = models.JSONField()

    class Meta:
        db_table = settings.BASE_DB_TABLE + '_customer'
        verbose_name = settings.BASE_DB_TABLE + ' Customer'
        verbose_name_plural = verbose_name
        ordering = ['-id']


class ASN(CoreModel):
    data = models.JSONField()

    class Meta:
        db_table = settings.BASE_DB_TABLE + '_asn'
        verbose_name = settings.BASE_DB_TABLE + ' ASN'
        verbose_name_plural = verbose_name
        ordering = ['-id']


class DN(CoreModel):
    data = models.JSONField()

    class Meta:
        db_table = settings.BASE_DB_TABLE + '_dn'
        verbose_name = settings.BASE_DB_TABLE + ' DN'
        verbose_name_plural = verbose_name
        ordering = ['-id']


class Purchase(CoreModel):
    data = models.JSONField()

    class Meta:
        db_table = settings.BASE_DB_TABLE + '_purchase'
        verbose_name = settings.BASE_DB_TABLE + ' Purchase'
        verbose_name_plural = verbose_name
        ordering = ['-id']


class Bar(CoreModel):
    data = models.JSONField()

    class Meta:
        db_table = settings.BASE_DB_TABLE + '_bar'
        verbose_name = settings.BASE_DB_TABLE + ' Bar'
        verbose_name_plural = verbose_name
        ordering = ['-id']


class Fee(CoreModel):
    data = models.JSONField()

    class Meta:
        db_table = settings.BASE_DB_TABLE + '_fee'
        verbose_name = settings.BASE_DB_TABLE + ' Fee'
        verbose_name_plural = verbose_name
        ordering = ['-id']


class Driver(CoreModel):
    data = models.JSONField()

    class Meta:
        db_table = settings.BASE_DB_TABLE + '_driver'
        verbose_name = settings.BASE_DB_TABLE + ' Driver'
        verbose_name_plural = verbose_name
        ordering = ['-id']


class PyPi(CoreModel):
    category = models.CharField(max_length=255, verbose_name="Category")
    date = models.DateTimeField(auto_now_add=False, auto_now=False, blank=True, verbose_name="PyPi Date")
    percent = models.FloatField(default=0, verbose_name="Percent")
    downloads = models.IntegerField(default=0, verbose_name="Downloads")

    class Meta:
        db_table = settings.BASE_DB_TABLE + '_pypi'
        verbose_name = settings.BASE_DB_TABLE + ' PyPi'
        verbose_name_plural = verbose_name
        ordering = ['-id']



import inspect
import time
from threading import Thread, Lock
import json
import importlib

from django.conf import settings
from django.core.cache import cache
from django_apscheduler.models import DjangoJob
from apscheduler.schedulers.background import BackgroundScheduler
from apscheduler.executors.pool import ThreadPoolExecutor
from django_apscheduler.jobstores import DjangoJobStore, register_events
from django_apscheduler.models import DjangoJobExecution
from datetime import datetime, timedelta
from bomiot.server.core.models import JobList

# map the args
ARGS_MAP = {
    'cron': ['year', 'month', 'day', 'week', 'day_of_week', 'hour', 'minute', 'second', 'start_date', 'end_date', 'timezone'],
    'interval': ['weeks', 'days', 'hours', 'minutes', 'seconds', 'start_date', 'end_date', 'timezone'],
    'date': ['run_date', 'timezone']
}

TIMEZONE = settings.TIME_ZONE if hasattr(settings, 'TIME_ZONE') else 'UTC'
executors = {
    'default': ThreadPoolExecutor(20),
}

scheduler = BackgroundScheduler(timezone=TIMEZONE, executors=executors)
scheduler.add_jobstore(DjangoJobStore(), 'default')
scheduler_lock = Lock()

class SchedulerManager(Thread):
    """
    Scheduler of Job
    """

    def __init__(self, scheduler):
        super(SchedulerManager, self).__init__()
        self.scheduler = scheduler
        self._stop_event = False
        register_events(self.scheduler)
        self.daemon = True
        JobList.objects.filter().delete()
        self.scheduler.start()
    
    def get_existing_jobs(self):
        try:
            return DjangoJob.objects.values_list('id', flat=True)
        except Exception as e:
            return []

    def get_active_jobs(self):
        try:
            return JobList.objects.filter(type=True)
        except Exception as e:
            return []

    def sync_jobs(self, force=False):
        with scheduler_lock:
            try:
                active_jobs = self.get_active_jobs()
                active_job_ids = {job.job_id for job in active_jobs}
                scheduled_job_ids = {job.id for job in self.scheduler.get_jobs()}
                for job_id in scheduled_job_ids.difference(active_job_ids):
                    try:
                        self.scheduler.remove_job(job_id)
                    except Exception as e:
                        print(f"Error removing job {job_id}: {str(e)}")
                scheduled_job_ids_res = {job.id for job in self.scheduler.get_jobs()}
                for job in active_job_ids.difference(scheduled_job_ids_res):
                    self._update_job(active_jobs.filter(job_id=job).first(), force)
                self.delete_old_job_executions()
            except Exception as e:
                print(f"Error syncing jobs: {str(e)}")

    def delete_old_job_executions(self, max_age=7):
        cutoff = datetime.now() - timedelta(days=max_age)
        old_executions = DjangoJobExecution.objects.filter(
            run_time__lt=cutoff
        )
        if old_executions.count() > 0:
            old_executions.delete()

    def _update_job(self, job, force=False):
        try:
            config = json.loads(job.configuration)
            trigger_type = job.trigger
            if trigger_type not in ARGS_MAP:
                return
            trigger_args = {
                arg: config.get(arg) 
                for arg in ARGS_MAP[trigger_type] 
                if config.get(arg) is not None
            }
            module = importlib.import_module(job.module_name)
            job_func = getattr(module, job.func_name)
            if not callable(job_func):
                return
            self.scheduler.add_job(
                func=job_func,
                trigger=trigger_type,
                id=job.job_id,
                replace_existing=True,
                kwargs={'sender': job_func, **trigger_args}
            )
        except Exception as e:
            job.type = False
            job.save()

    def run(self):
        try:
            self.sync_jobs(force=True)
            # print("Scheduler manager started")
            while not self._stop_event:
                self.sync_jobs()
                time.sleep(60)                
        except Exception as e:
            print(f"Scheduler manager crashed: {str(e)}")
            
    def stop(self):
        self._stop_event = True
        self.scheduler.shutdown()
        print("Scheduler manager stopped")


# init scheduler manager
sm = SchedulerManager(scheduler)



from os.path import join, exists
from os import makedirs, getcwd, rename
import shutil
from pathlib import Path
from .init import create_file
from configparser import ConfigParser
import sys
import os
import importlib.resources

def deploy(folder: str):
    """
    deploy project
    :param folder:
    :return:
    """

    if len(sys.argv) < 3:
        print('Please enter your deploy project name')
    else:
        create_file('')
        exists(join(getcwd(), 'deploy')) or makedirs(join(getcwd(), 'deploy'))
        deploy_path = join(getcwd(), 'deploy')
        if str(sys.argv[2]) != 'bomiot':
            supervisor_path = join(deploy_path, f'{sys.argv[2]}.conf')
            if exists(supervisor_path) is False:
                with open(supervisor_path, "wt") as f:
                    f.write("")
                supervisor_config = ConfigParser()
                supervisor_config.read(supervisor_path)
                supervisor_config.add_section(f'program:{sys.argv[2]}')
                supervisor_config.set(f'program:{sys.argv[2]}', 'user', 'root')
                supervisor_config.set(f'program:{sys.argv[2]}', 'command', 'bomiot run -b 0.0.0.0 -p 8000 -w 4')
                supervisor_config.set(f'program:{sys.argv[2]}', 'autostart', 'true')
                supervisor_config.set(f'program:{sys.argv[2]}', 'autorestart', 'true')
                supervisor_config.set(f'program:{sys.argv[2]}', 'startretries', '3')
                supervisor_config.set(f'program:{sys.argv[2]}', 'exitcodes', '0,2')
                supervisor_config.set(f'program:{sys.argv[2]}', 'stopsignal', 'TERM')
                supervisor_config.set(f'program:{sys.argv[2]}', 'stopwaitsecs', '10')
                supervisor_config.set(f'program:{sys.argv[2]}', 'stopasgroup', 'true')
                supervisor_config.set(f'program:{sys.argv[2]}', 'killasgroup', 'true')
                supervisor_config.set(f'program:{sys.argv[2]}', 'directory', f'{getcwd()}')
                supervisor_config.set(f'program:{sys.argv[2]}', 'stdout_logfile',
                        f'{join(getcwd(), "logs", f"{sys.argv[2]}_access.log")}')
                supervisor_config.set(f'program:{sys.argv[2]}', 'stderr_logfile',
                        f'{join(getcwd(), "logs", f"{sys.argv[2]}_error.log")}')
                supervisor_config.set(f'program:{sys.argv[2]}', 'stdout_logfile_maxbytes', '20MB')
                supervisor_config.set(f'program:{sys.argv[2]}', 'stdout_logfile_backups', '10')
                supervisor_config.set(f'program:{sys.argv[2]}', 'stderr_logfile_maxbytes', '20MB')
                supervisor_config.set(f'program:{sys.argv[2]}', 'stderr_logfile_backups', '10')

                supervisor_config.write(open(supervisor_path, "wt"))


        print(f'Deploy project {str(sys.argv[2])} workspace success')



from rest_framework.views import exception_handler
from rest_framework.response import Response
from django.db import DatabaseError


def custom_exception_handler(exc, context):
    """
    custom exception handler for DRF
    :param exc: exception 
    :param context: context information
    :return: Response
    """
    # use DRF's default exception handler to get the standard error response
    response = exception_handler(exc, context)

    if response is not None:
        # add additional information to the response
        response.data['status_code'] = response.status_code
        response = Response(response.data)
    else:
        # handle the exception
        print(exc, DatabaseError)
        if isinstance(exc, DatabaseError):
            response = Response({'msg': 'A database error occurred.'})
        else:
            # handle other exceptions
            # response = Response({'msg': 'An unknown error occurred.'})
            pass

    return response